##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
from pjpersist import testing, mquery


def run(sqlx):
    print(sqlx.__sqlrepr__('postgres'))


def test_convert():
    r"""Test mquery.convert.

    We can parse MongoDB queries and convert them to SQLBuilder
    expressions.  First we inititalize the converter with our table
    and field names:

        >>> mq = mquery.Converter("Bar", "data")

    Let's enable a mode that simplifies some crud for nicer tests:

        >>> mq.simplified = True

    Now we can run a basic query:

        >>> mq.convert({'day': 'Friday', 'drink': 'whiskey', 'nr': 42})
        <SQLOp ...>

        >>> run(_)
        ((((Bar.data) -> ('day')) = ('"Friday"')) AND
         ((((Bar.data) -> ('drink')) = ('"whiskey"')) AND
          (((Bar.data) -> ('nr')) = ('42'))))

    Actually, this is wrong, because in mongo the test could as well
    mean checking for an member in an array:

        >>> mq.simplified = False
        >>> run(mq.convert({'nr': 42}))
        ((((Bar.data) -> ('nr')) = ('42')) OR
         (('[42]'::jsonb) <@ ((Bar.data) -> ('nr'))))

    Enough craziness:

        >>> mq.simplified = True

    Keys can be dotted paths:

        >>> run(mq.convert({'contract.week.day': 'Friday'}))
        (((Bar.data) #> (array['contract', 'week', 'day'])) = ('"Friday"'))

    Values can be objects with comparison directives:

        >>> run(mq.convert({'quantity': {'$gt': 20}}))
        (((Bar.data) -> ('quantity')) > ('20'))

        >>> run(mq.convert({'quantity': {'$lte': 20}}))
        (((Bar.data) -> ('quantity')) <= ('20'))

        >>> run(mq.convert({'quantity': {'$in': [20, 30, 40]}}))
        ((((Bar.data) -> ('quantity')) IN
                  ('20'::jsonb, '30'::jsonb, '40'::jsonb)) OR
         ((((Bar.data) -> ('quantity')) @> ('20')) OR
         ((((Bar.data) -> ('quantity')) @> ('30')) OR
         (((Bar.data) -> ('quantity')) @> ('40')))))

        >>> run(mq.convert({'quantity': {'$nin': [20, 30, 40]}}))
        NOT ((((Bar.data) -> ('quantity')) IN
                      ('20'::jsonb, '30'::jsonb, '40'::jsonb)) OR
             ((((Bar.data) -> ('quantity')) @> ('20')) OR
             ((((Bar.data) -> ('quantity')) @> ('30')) OR
             (((Bar.data) -> ('quantity')) @> ('40')))))


    The $not operator.  It matches if the field does not exist or
    does not match the condition:

        >>> run(mq.convert({'quantity': {'$not': {'$gt': 20}}}))
        ((((Bar.data) -> ('quantity')) IS NULL) OR
         (NOT (((Bar.data) -> ('quantity')) > ('20'))))

    The $size operator:

        >>> run(mq.convert({'quantities': {'$size': 3}}))
        ((jsonb_array_length(((Bar.data) -> ('quantities')))) = (3))

        >>> run(mq.convert({'some.quantities': {'$size': 3}}))
        ((jsonb_array_length(((Bar.data) #> (array['some', 'quantities'])))) = (3))

    The $exists operator

        >>> run(mq.convert({'quantities': {'$exists': True}}))
        (((Bar.data) -> ('quantities')) IS NOT NULL)

        >>> run(mq.convert({'quantities': {'$exists': False}}))
        (((Bar.data) -> ('quantities')) IS NULL)

    There can be several operators in one dict:

        >>> run(mq.convert({'qty': {'$exists': True, '$nin': [1, 2]}}))
        ((((Bar.data) -> ('qty')) IS NOT NULL) AND
         (NOT ((((Bar.data) -> ('qty')) IN ('1'::jsonb, '2'::jsonb)) OR
              ((((Bar.data) -> ('qty')) @> ('1')) OR
               (((Bar.data) -> ('qty')) @> ('2'))))))

    The $all operator

        >>> run(mq.convert({'some.quantities': {'$all': [1, 3]}}))
        (((Bar.data) #> (array['some', 'quantities'])) @> ('[1, 3]'))

    $elemMatch is tricky:

        >>> run(mq.convert({'nrs': {'$elemMatch': [{'$gt': 2}, {'$lte': 3}]}}))
        EXISTS (SELECT value FROM jsonb_array_elements(((Bar.data) -> ('nrs')))
                WHERE (((value) > ('2')) AND
                       ((value) <= ('3'))))

    $startswith is to replace a $regex:

        >>> run(mq.convert({'drinks': {'$startswith': 'good'}}))
        (((Bar.data) -> ('drinks'))::text LIKE ('good%') ESCAPE E'\\')

    """


def test_convert_logical():
    """Test mquery.convert logical operators.

    Setup:

        >>> mq = mquery.Converter("Bar", "data")
        >>> mq.simplified = True

    There are three operators that aggregate  lists of queries:

        >>> run(mq.convert({
        ...     '$and': [{'mode': 1}, {'foo': 'bar'}]}))
        ((((Bar.data) -> ('mode')) = ('1')) AND
         (((Bar.data) -> ('foo')) = ('"bar"')))

        >>> run(mq.convert({
        ...     '$or': [{'mode': 1}, {'foo': 'bar'}]}))
        ((((Bar.data) -> ('mode')) = ('1')) OR (((Bar.data) -> ('foo')) = ('"bar"')))

        >>> run(mq.convert({
        ...     '$nor': [{'mode': 1}, {'foo': 'bar'}]}))
        NOT ((((Bar.data) -> ('mode')) = ('1')) OR (((Bar.data) -> ('foo')) = ('"bar"')))

    """


def test_convert_id():
    """Test mquery.convert _id handling.

    Setup:

        >>> mq = mquery.Converter("Bar", "data")
        >>> mq.simplified = True

    Fields named _id are translated to query the id column directly on
    the table:

        >>> run(mq.convert({'_id': '3334444555', "lastname": "Getinthechopper"}))
        (((Bar.id) = ('3334444555')) AND
         (((Bar.data) -> ('lastname')) = ('"Getinthechopper"')))

        >>> run(mq.convert({'_id': {'$exists': True}}))
        ((Bar.id) IS NOT NULL)

    """


def test_convert_datetime():
    """Test mquery.convert datetime handling.

    Setup:

        >>> mq = mquery.Converter("Bar", "data")
        >>> mq.simplified = True
        >>> import datetime
        >>> testing.setUpSerializers(None)

    datetime fields require special serialization:

        >>> run(mq.convert(
        ...     {'ts': datetime.datetime(2014,12,11,11,12,14),
        ...      'dte': datetime.date(2014,12,11)}))
        ((((Bar.data) -> ('dte')) = ('{"_py_type": "datetime.date", "value": "2014-12-11"}'))
        AND (((Bar.data) -> ('ts')) = ('{"_py_type": "datetime.datetime", "value": "2014-12-11T11:12:14.000000"}')))

        >>> testing.tearDownSerializers(None)

    """


def test_suite():
    return doctest.DocTestSuite(
        optionflags=testing.OPTIONFLAGS)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Statistics on executed queries"""
import sys
from collections import namedtuple

from zope.exceptions import exceptionformatter


QueryStats = namedtuple("QueryStats",
                        ["query", "args", "time", "traceback", "database"])

QueryTotals = namedtuple('QueryTotals',
                         ["total_queries", "total_time", "sorted_queries"])


# Number of most expensive queries to print out
NUM_OF_QUERIES_TO_REPORT = 10

REPORT_TRACEBACK = False

# Traceback limit
TB_LIMIT = 15  # 15 should be sufficient to figure


class QueryReport(object):
    def __init__(self):
        self.qlog = []
        self.report_traceback = REPORT_TRACEBACK

    def record(self, query, args, elapsed_time, database=None):
        """Record executed query

        elapsed_time is time, elapsed by executing query, in secodes
        """
        traceback = self._collect_traceback()
        self.qlog.append(QueryStats(query, args, elapsed_time,
                                    traceback, database))

    def calc_totals(self):
        """Calculate totals and return QueryTotals object
        """

        total_queries = len(self.qlog)
        total_time = sum(q.time for q in self.qlog)
        sorted_queries = sorted(self.qlog, key=lambda q: q.time)
        return QueryTotals(total_queries, total_time, sorted_queries)

    def calc_and_report(self):
        """Calculate totals and print out report
        """
        if len(self.qlog) == 0:
            return "Query report: no queries were executed"
        totals = self.calc_totals()
        sep = '-' * 60

        report = []
        p = report.append

        p("Query report:")
        p(sep)
        p("%s most expensive queries:" % NUM_OF_QUERIES_TO_REPORT)
        for q in totals.sorted_queries[-NUM_OF_QUERIES_TO_REPORT:]:
            p("*** %s" % q.query)
            p("... ARGS: %s" % (q.args,))
            p("... TIME: %.4fms" % (q.time * 1000))
            if self.report_traceback and q.traceback:
                p(q.traceback)
            p("")
        p(sep)
        p("Queries executed: %s" % totals.total_queries)
        p("Time spent: %.4fms" % (totals.total_time * 1000))

        return "\n".join(report)

    def clear(self):
        self.qlog = []

    def _collect_traceback(self):
        try:
            raise ValueError('boom')
        except:
            # we need here exceptionformatter, otherwise __traceback_info__
            # is not added
            stack = exceptionformatter.extract_stack(
                sys.exc_info()[2].tb_frame.f_back, limit=TB_LIMIT)
            tb = ''.join(stack[:-2])
            return tb



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Mongo Persistence Testing Support"""
import doctest
import io
import logging
import os
import psycopg2
import psycopg2.extras
import psycopg2.pool
import re
import sys
import threading
import transaction
import unittest
from pprint import pprint

import zope.component
from zope.testing import module, renormalizing

from pjpersist import datamanager, serialize, interfaces

py3checkers = [
    # Mangle unicode strings
    (re.compile("u('.*?')"), r"\1"),
    (re.compile('u(".*?")'), r"\1"),
    # Mangle long ints
    (re.compile('([0-9]+)L$'), r"\1"),
    (re.compile('__builtin__'), 'builtins'),
    (re.compile('pjpersist.interfaces.CircularReferenceError'),
     'CircularReferenceError'),
]

checker = renormalizing.RENormalizing([
    # Date/Time objects
    (re.compile(r'datetime.datetime\(.*\)'),
     'datetime.datetime(2011, 10, 1, 9, 45)'),
    # IDs
    (re.compile(r"'[0-9a-f]{24}'"), "'0001020304050607080a0b0c0'"),
    ] + py3checkers)

OPTIONFLAGS = (
    doctest.NORMALIZE_WHITESPACE|
    doctest.ELLIPSIS|
    doctest.REPORT_ONLY_FIRST_FAILURE
    #|doctest.REPORT_NDIFF
    )

DBNAME = 'pjpersist_test'
DBNAME_OTHER = 'pjpersist_test_other'


class DummyConnectionPool:
    def __init__(self, conn):
        self._available = conn
        self._taken = None

    def getconn(self):
        if self._available is None:
            raise psycopg2.pool.PoolError("Connection is already taken")
        self._available.reset()
        self._available.set_isolation_level(
            psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE)
        self._taken = self._available
        self._available = None
        return self._taken

    def putconn(self, conn, key=None, close=False):
        assert conn is self._taken
        self._available = self._taken
        self._taken = None

    def isTaken(self):
        return self._taken is not None


@zope.interface.implementer(interfaces.IPJDataManagerProvider)
class SimpleDataManagerProvider(object):
    def __init__(self, dms, default=None):
        self.idx = {dm.database: dm for dm in dms}
        self.idx[None] = default

    def get(self, database):
        return self.idx[database]


def getConnection(database=None):
    conn = psycopg2.connect(
        database=database or 'template1',
        host='localhost', port=5432,
        user='pjpersist', password='pjpersist')
    conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE)
    return conn


def createDB():
    dropDB()
    conn = getConnection()
    with conn.cursor() as cur:
        cur.execute('END')
        cur.execute('DROP DATABASE IF EXISTS %s' % DBNAME)
        cur.execute('CREATE DATABASE %s' % DBNAME)
        cur.execute('DROP DATABASE IF EXISTS %s' % DBNAME_OTHER)
        cur.execute('CREATE DATABASE %s' % DBNAME_OTHER)
    conn.commit()
    conn.close()


def dropDB():
    conn = getConnection()
    with conn.cursor() as cur:
        cur.execute('END')
        try:
            cur.execute('DROP DATABASE IF EXISTS %s' % DBNAME_OTHER)
            cur.execute('DROP DATABASE IF EXISTS %s' % DBNAME)
        except psycopg2.ProgrammingError:
            pass
    conn.commit()
    conn.close()


def cleanDB(conn=None):
    if conn is None:
        conn = getConnection(DBNAME)
    conn.rollback()
    with conn.cursor() as cur:
        cur.execute("""SELECT tablename FROM pg_tables""")
        for res in cur.fetchall():
            if not res[0].startswith('pg_') and not res[0].startswith('sql_'):
                cur.execute('DROP TABLE ' + res[0])
    conn.commit()


def setUpSerializers(test):
    serialize.SERIALIZERS = []


def tearDownSerializers(test):
    del serialize.SERIALIZERS[:]


def setUp(test):
    module.setUp(test)
    setUpSerializers(test)
    g = test.globs
    g['conn'] = getConnection(DBNAME)
    g['conn_other'] = getConnection(DBNAME_OTHER)
    cleanDB(g['conn'])
    cleanDB(g['conn_other'])
    g['commit'] = transaction.commit
    g['dm'] = datamanager.PJDataManager(DummyConnectionPool(g['conn']))
    g['dm_other'] = datamanager.PJDataManager(DummyConnectionPool(g['conn_other']))

    def dumpTable(table, flush=True, isolate=False):
        if isolate:
            conn = getConnection(database=DBNAME)
        else:
            conn = g['conn']
        with conn.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
            try:
                cur.execute('SELECT * FROM ' + table)
            except psycopg2.ProgrammingError as err:
                print(err)
            else:
                pprint([dict(e) for e in cur.fetchall()])
        if isolate:
            conn.close()
    g['dumpTable'] = dumpTable

    dmp = SimpleDataManagerProvider([g['dm'], g['dm_other']], g['dm'])
    zope.component.provideUtility(dmp)


def tearDown(test):
    module.tearDown(test)
    tearDownSerializers(test)
    transaction.abort()
    cleanDB(test.globs['conn'])
    cleanDB(test.globs['conn_other'])
    test.globs['conn'].close()
    test.globs['conn_other'].close()
    resetCaches()


class DatabaseLayer(object):
    __bases__ = ()

    def __init__(self, name):
        self.__name__ = name

    def setUp(self):
        createDB()
        self.setUpSqlLogging()

    def tearDown(self):
        self.tearDownSqlLogging()
        dropDB()

    def setUpSqlLogging(self):
        if "SHOW_SQL" not in os.environ:
            return

        self.save_PJ_ACCESS_LOGGING = datamanager.PJ_ACCESS_LOGGING
        datamanager.PJ_ACCESS_LOGGING = True

        setUpLogging(datamanager.TABLE_LOG, copy_to_stdout=True)
        setUpLogging(datamanager.LOG, copy_to_stdout=True)

    def tearDownSqlLogging(self):
        if "SHOW_SQL" not in os.environ:
            return

        tearDownLogging(datamanager.LOG)
        tearDownLogging(datamanager.TABLE_LOG)

        datamanager.PJ_ACCESS_LOGGING = self.save_PJ_ACCESS_LOGGING


db_layer = DatabaseLayer("db_layer")


class PJTestCase(unittest.TestCase):
    layer = db_layer

    def setUp(self):
        setUpSerializers(self)
        self.conn = getConnection(DBNAME)
        cleanDB(self.conn)
        self.dm = datamanager.PJDataManager(DummyConnectionPool(self.conn))

    def tearDown(self):
        datamanager.CONFLICT_TRACEBACK_INFO.traceback = None
        tearDownSerializers(self)
        transaction.abort()
        cleanDB(self.conn)
        self.conn.close()
        resetCaches()


def resetCaches():
    serialize.AVAILABLE_NAME_MAPPINGS.__init__()
    serialize.PATH_RESOLVE_CACHE = {}
    serialize.TABLE_KLASS_MAP = {}


def log_sql_to_file(fname, add_tb=True, tb_limit=15):
    import logging

    datamanager.PJ_ENABLE_QUERY_STATS = True
    datamanager.PJ_ACCESS_LOGGING = True
    datamanager.TABLE_LOG.setLevel(logging.DEBUG)
    datamanager.PJPersistCursor.TB_LIMIT = tb_limit

    fh = logging.FileHandler(fname)
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        '%(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    datamanager.TABLE_LOG.addHandler(fh)


class StdoutHandler(logging.StreamHandler):
    """Logging handler that follows the current binding of sys.stdout."""

    def __init__(self):
        # skip logging.StreamHandler.__init__()
        logging.Handler.__init__(self)

    @property
    def stream(self):
        return sys.stdout


def setUpLogging(logger, level=logging.DEBUG, format='%(message)s',
                 copy_to_stdout=False):
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    buf = io.StringIO()
    handler = logging.StreamHandler(buf)
    handler._added_by_tests_ = True
    handler._old_propagate_ = logger.propagate
    handler._old_level_ = logger.level
    handler.setFormatter(logging.Formatter(format))
    logger.addHandler(handler)
    if copy_to_stdout:
        # can't use logging.StreamHandler(sys.stdout) because sys.stdout might
        # be changed latter to a StringIO, and we want messages to be seen
        # by doctests.
        handler = StdoutHandler()
        handler._added_by_tests_ = True
        handler._old_propagate_ = logger.propagate
        handler._old_level_ = logger.level
        handler.setFormatter(logging.Formatter(format))
        logger.addHandler(handler)
    logger.propagate = False
    logger.setLevel(level)
    return buf


def tearDownLogging(logger):
    if isinstance(logger, str):
        logger = logging.getLogger(logger)
    for handler in list(logger.handlers):
        if hasattr(handler, '_added_by_tests_'):
            logger.removeHandler(handler)
            logger.propagate = handler._old_propagate_
            logger.setLevel(handler._old_level_)


#TO_JOIN = []
def run_in_thread(func):
    t = threading.Thread(target=func)
    t.setDaemon(True)
    t.start()
    #TO_JOIN.append(t)
    return t



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Persistence Zope Container Interfaces"""
import zope.interface
import zope.schema


class IPJContainer(zope.interface.Interface):

    _pj_table = zope.schema.ASCIILine(
        title='PostGreSQL Table',
        description=(
            'Specifies the PostGreSQL table in which to store items.')
        )

    _pj_mapping_key = zope.schema.ASCIILine(
        title='Mapping Key',
        description=(
            'Specifies the attribute name of the item that is used as the '
            'mapping/dictionary/container key.'),
        default='key')

    _pj_parent_key = zope.schema.ASCIILine(
        title='Parent Key',
        description=(
            'Specifies the attribute name of the item that is used to store '
            'the parent/container reference.'),
        default='parent')

    _pj_remove_documents = zope.schema.Bool(
        title='Remove Documents',
        description=(
            'A flag when set causes documents to be removed from the DB when '
            'they are removed from the container.'),
        default=True)

    _pj_column_fields = zope.schema.Tuple(
        title='Column Fields',
        description=(
            'A list of field NAMES that represent columns '
            'in the storage table. the default values are `id` and `data`, '
            'but more can be added.'
            ''
            'These fields are used to figure whether to take native SQL columns '
            'instead of JSONB fields in `raw_find` '
            'see also pjpersist.persistent.SimpleColumnSerialization._pj_column_fields'
            ''
            'NOTE: this is a list of NAMES (str) opposed to '
            'IColumnSerialization._pj_column_fields are zope.schema fields!'),
        default=('id', 'data'),
        required=True)

    def _pj_get_parent_key_value():
        """Returns the value that is used to specify a particular container as
        the parent of the item.
        """

    def _pj_get_resolve_filter():
        """Returns a query to apply when querying for single id
        """

    def _pj_get_list_filter():
        """Returns a query spec representing a filter that only returns
        objects in this container."""

    def convert_mongo_query(spec):
        """BBB: providing support for mongo style queries"""

    def raw_find(qry, fields=(), **kwargs):
        """Return a raw psycopg result cursor for the specified query.

        The qry is updated to also contain the container's filter condition.
        ``kwargs`` allows you to pass parameters to sqlbuilder.Select

        Note: The user is responsible of closing the cursor after use.
        """

    def find(qry, **kwargs):
        """Return a Python object result set for the specified query.

        The qry is updated to also contain the container's filter condition.
        ``kwargs`` allows you to pass parameters to sqlbuilder.Select

        Note: The user is responsible of closing the cursor after use.
        """

    def raw_find_one(qry=None, id=None):
        """Return the record for the specified query.

        At least one of the arguments must be specified. If no document was
        found, returns None.

        The qry is updated to also contain the container's filter condition.

        Note: The user is responsible of closing the cursor after use.
        """

    def find_one(qry=None, id=None):
        """Return a single Python object for the specified query.

        At least one of the arguments must be specified.

        The qry is updated to also contain the container's filter condition.

        Note: The user is responsible of closing the cursor after use.
        """

    def count(qry=None):
        """Count items matching query conditions

        Provides a quick way to count items without fetching them from
        database.
        """

    def add(value, key=None):
        """Add an object without necessarily knowing the key of the object.

        Either pass in a valid key or the key will be:
        - in case ``_pj_mapping_key`` is ``None``: the object's OID
        - otherwise ``getattr(value, _pj_mapping_key)``
        """

    def clear():
        """Delete all items from this container.

        Note, that this will not touch all items from the collection, but only
        those, specified in ``_pj_get_items_filter``.
        """



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Object Serialization for PostGreSQL's JSONB"""
import base64
import copyreg
import datetime
import warnings

import persistent.interfaces
import persistent.dict
import persistent.list
import zope.interface
from zope.dottedname.resolve import resolve
from decimal import Decimal

from pjpersist import interfaces

ALWAYS_READ_FULL_DOC = True

SERIALIZERS = []
AVAILABLE_NAME_MAPPINGS = set()
PATH_RESOLVE_CACHE = {}
TABLE_KLASS_MAP = {}

FMT_DATE = "%Y-%m-%d"
FMT_TIME = "%H:%M:%S.%f"
FMT_DATETIME = "%Y-%m-%dT%H:%M:%S.%f"

# BBB: Will be removed in 2.0.
FMT_TIME_BBB = "%H:%M:%S"
FMT_DATETIME_BBB = "%Y-%m-%dT%H:%M:%S"

# actually we should extract this somehow from psycopg2
PYTHON_TO_PG_TYPES = {
    str: "text",
    bytes: "text",
    bool: "bool",
    float: "double",
    int: "integer",
    Decimal: "numeric",
    datetime.date: "date",
    datetime.time: "time",
    datetime.datetime: "timestamptz",
    datetime.timedelta: "interval",
    list: "array",
}


KNOWN_FACTORIES = {
    '__builtin__.set': set,
}

# we need to convert dicts and dict-ish values which have non-string keys
# to a structure that can be stored with JSONB (which accepts only strings
# as keys)
# so we do:
# A) happy day
# {'1': 1, '2': '2', '3': Number(3)}
# ->
# {'1': 1, '2': '2', '3': {'_py_type': '__main__.Number', 'num': 3}}
#
# B) if there's a non string key:
# {1: 'one', 2: 'two', 3: 'three'}
# ->
# {DICT_NON_STRING_KEY_MARKER: [(1, 'one'), (2, 'two'), (3, 'three')]}
#
# C) if there's `dict_data` in the input:
# {'1': 1, 'dict_data': 'works?'}
# ->
# {DICT_NON_STRING_KEY_MARKER: [('1', 1), ('dict_data', 'works?')]}
#
# case C is required otherwise when converting back to a dict by
# `ObjectReader.get_object` `get_object` would fail because it expects
# our converted list-in-a-dict structure
# basically we do not allow the DICT_NON_STRING_KEY_MARKER to pass through
DICT_NON_STRING_KEY_MARKER = 'dict_data'


def get_dotted_name(obj, escape=False):
    name = obj.__module__ + '.' + obj.__name__
    if not escape:
        return name
    # Make the name safe.
    name = name.replace('.', '_dot_')
    # XXX: Circumventing a bug in sqlobject.sqlbuilder that prohibits names to
    # start with _.
    name = 'u'+name if name.startswith('_') else name
    return name


def link_to_parent(obj, pobj):
    if obj._p_jar is None:
        if pobj is not None and  getattr(pobj, '_p_jar', None) is not None:
            obj._p_jar = pobj._p_jar
        setattr(obj, interfaces.DOC_OBJECT_ATTR_NAME, pobj)


class PersistentDict(persistent.dict.PersistentDict):
    # SUB_OBJECT_ATTR_NAME:
    _p_pj_sub_object = True

    def __init__(self, data=None, **kwargs):
        # We optimize the case where data is not a dict. The original
        # implementation always created an empty dict, which it then
        # updated. This turned out to be expensive.
        if data is None:
            self.data = {}
        elif isinstance(data, dict):
            self.data = data.copy()
        else:
            self.data = dict(data)
        if len(kwargs):
            self.update(kwargs)

    def __getitem__(self, key):
        # The UserDict supports a __missing__() function, which I have never
        # seen or used before, but it makes the method significantly
        # slower. So let's not do that.
        return self.data[key]

    def __eq__(self, other):
        return self.data == other

    def __ne__(self, other):
        return not self.__eq__(other)


class PersistentList(persistent.list.PersistentList):
    # SUB_OBJECT_ATTR_NAME:
    _p_pj_sub_object = True


class DBRef(object):

    def __init__(self, table, id, database=None):
        self._table = table
        self._id = id
        self._database = database
        self.__calculate_hash()

    def __calculate_hash(self):
        self.hash = hash(str(self.database)+str(self.table)+str(self.id))

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        self._database = value
        self.__calculate_hash()

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, value):
        self._table = value
        self.__calculate_hash()

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        self._id = value
        self.__calculate_hash()

    def __setstate__(self, state):
        self.__init__(state['table'], state['id'], state['database'])

    def __getstate__(self):
        return {'database': self.database,
                'table': self.table,
                'id': self.id}

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        try:
            return self.hash == other.hash
        except AttributeError:
            # `other` is not a DBRef or is None
            return False

    def __repr__(self):
        return 'DBRef(%r, %r, %r)' %(self.table, self.id, self.database)

    def as_tuple(self):
        return self.database, self.table, self.id

    def as_json(self):
        return {'_py_type': 'DBREF',
                'database': self.database,
                'table': self.table,
                'id': self.id}


class Binary(str):
    pass

@zope.interface.implementer(interfaces.IObjectSerializer)
class ObjectSerializer(object):

    def can_read(self, state):
        raise NotImplementedError

    def read(self, state):
        raise NotImplementedError

    def can_write(self, obj):
        raise NotImplementedError

    def write(self, obj):
        raise NotImplementedError


@zope.interface.implementer(interfaces.IObjectWriter)
class ObjectWriter(object):

    def __init__(self, jar):
        self._jar = jar

    def get_table_name(self, obj):
        __traceback_info__ = obj
        db_name = getattr(
            obj, interfaces.DATABASE_ATTR_NAME,
            self._jar.database if self._jar else None)
        try:
            table_name = getattr(obj, interfaces.TABLE_ATTR_NAME)
        except AttributeError:
            return db_name, get_dotted_name(obj.__class__, True)
        return db_name, table_name

    def get_non_persistent_state(self, obj):
        objectId = id(obj)
        objectType = type(obj)
        __traceback_info__ = obj, objectType, objectId
        # XXX: Look at the pickle library how to properly handle all types and
        # old-style classes with all of the possible pickle extensions.

        # Get the state of the object. Only pickable objects can be reduced.
        reduce_fn = copyreg.dispatch_table.get(objectType)
        if reduce_fn is not None:
            reduced = reduce_fn(obj)
        else:
            # XXX: __reduce_ex__
            reduced = obj.__reduce__()
        # The full object state (item 3) seems to be optional, so let's make
        # sure we handle that case gracefully.
        if isinstance(reduced, str):
            # When the reduced state is just a string it represents a name in
            # a module. The module will be extracted from __module__.
            return {'_py_constant': obj.__module__+'.'+reduced}
        if len(reduced) == 2:
            factory, args = reduced
            obj_state = {}
        else:
            factory, args, obj_state = reduced
            if obj_state is None:
                obj_state = {}
        # We are trying very hard to create a clean JSONB (sub-)document. But
        # we need a little bit of meta-data to help us out later.
        if factory == copyreg._reconstructor and \
               args == (obj.__class__, object, None):
            # This is the simple case, which means we can produce a nicer
            # JSONB output.
            state = {'_py_type': get_dotted_name(args[0])}
        elif factory == copyreg.__newobj__ and args == (obj.__class__,):
            # Another simple case for persistent objects that do not want
            # their own document.
            state = {interfaces.PY_TYPE_ATTR_NAME: get_dotted_name(args[0])}
        else:
            state = {'_py_factory': get_dotted_name(factory),
                     '_py_factory_args': self.get_state(args, obj)}
        for name, value in obj_state.items():
            state[name] = self.get_state(value, obj)
        return state

    def get_persistent_state(self, obj):
        #seen.add(id(obj))
        __traceback_info__ = obj
        # Persistent sub-objects are stored by reference, the key being
        # (table name, oid).
        # Getting the table name is easy, but if we have an unsaved
        # persistent object, we do not yet have an OID. This must be solved by
        # storing the persistent object.
        if obj._p_oid is None:
            dbref = self.store(obj, ref_only=True)
        else:
            db_name, table_name = self.get_table_name(obj)
            dbref = obj._p_oid
        # Create the reference sub-document. The _p_type value helps with the
        # deserialization later.
        return dbref.as_json()

    def get_state(self, obj, pobj=None):
        objectType = type(obj)
        # in_seen = seen
        # if seen is None:
        #     seen = set()
        __traceback_info__ = obj, objectType, pobj
        if objectType in interfaces.PJ_NATIVE_TYPES:
            # If we have a native type, we'll just use it as the state.
            return obj
        if type(obj) == bytes:
            return {
                '_py_type': 'BINARY',
                'data': base64.b64encode(obj).decode('ascii')
            }
        if type(obj) == str:
            return obj
        # Some objects might not naturally serialize well and create a very
        # ugly JSONB entry. Thus, we allow custom serializers to be
        # registered, which can encode/decode different types of objects.
        for serializer in SERIALIZERS:
            if serializer.can_write(obj):
                return serializer.write(obj)

        if objectType == datetime.date:
            return {'_py_type': 'datetime.date',
                    'value': obj.strftime(FMT_DATE)}
        if objectType == datetime.time:
            return {'_py_type': 'datetime.time',
                    'value': obj.strftime(FMT_TIME)}
        if objectType == datetime.datetime:
            return {'_py_type': 'datetime.datetime',
                    'value': obj.strftime(FMT_DATETIME)}

        if isinstance(obj, type):
            # We frequently store class and function paths as meta-data, so we
            # need to be able to properly encode those.
            return {'_py_type': 'type',
                    'path': get_dotted_name(obj)}

        # We need to make sure that the object's jar and doc-object are
        # set. This is important for the case when a sub-object was just
        # added.
        if getattr(obj, interfaces.SUB_OBJECT_ATTR_NAME, False):
            if obj._p_jar is None:
                if pobj is not None and \
                        getattr(pobj, '_p_jar', None) is not None:
                    obj._p_jar = pobj._p_jar
                setattr(obj, interfaces.DOC_OBJECT_ATTR_NAME, pobj)

        if isinstance(obj, (tuple, list, PersistentList)):
            # Make sure that all values within a list are serialized
            # correctly. Also convert any sequence-type to a simple list.
            return [self.get_state(value, pobj) for value in obj]
        if isinstance(obj, (dict, PersistentDict)):
            # Same as for sequences, make sure that the contained values are
            # properly serialized.
            # Note: see comments at the definition of DICT_NON_STRING_KEY_MARKER
            has_non_compliant_key = False
            data = []
            for key, value in obj.items():
                data.append((key, self.get_state(value, pobj)))
                if (not isinstance(key, str) or  # non-string
                        # a key with our special marker
                        key==DICT_NON_STRING_KEY_MARKER):
                    has_non_compliant_key = True
            if not has_non_compliant_key:
                # The easy case: all keys are compliant:
                return dict(data)
            else:
                # We first need to reduce the keys and then produce a data
                # structure.
                data = [(self.get_state(key, pobj), value)
                        for key, value in data]
                return {DICT_NON_STRING_KEY_MARKER: data}

        if isinstance(obj, persistent.Persistent):
            # Only create a persistent reference, if the object does not want
            # to be a sub-document.
            if not getattr(obj, interfaces.SUB_OBJECT_ATTR_NAME, False):
                return self.get_persistent_state(obj)
            # This persistent object is a sub-document, so it is treated like
            # a non-persistent object.

        try:
            res = self.get_non_persistent_state(obj)
        except RuntimeError as re:
            # let it run into a RuntimeError...
            # it's hard to catch a non-persistent - non-persistent circular
            #    reference while NOT catching a
            #    >>> anobj = object()
            #    >>> alist = [anobj, anobj]

            if re.args[0].startswith('maximum recursion depth exceeded'):
                raise interfaces.CircularReferenceError(obj)
            else:
                raise
        return res

    def get_full_state(self, obj):
        doc = self.get_state(obj.__getstate__(), obj)
        # Always add a persistent type info
        doc[interfaces.PY_TYPE_ATTR_NAME] = get_dotted_name(obj.__class__)
        # Return the full state document
        return doc

    def store(self, obj, ref_only=False, id=None):
        __traceback_info__ = (obj, ref_only)

        # If it is the first time that this type of object is stored, getting
        # the table name has the side affect of telling the class whether it
        # has to store its Python type as well. So, do not remove, even if the
        # data is not used right away,
        db_name, table_name = self.get_table_name(obj)

        if obj._p_oid is None:
            # if not yet added,
            # _p_jar needs to be set early for self.get_state and subobjects
            # for subobjects _p_jar to be set
            obj._p_jar = self._jar

        if ref_only:
            # We only want to get OID quickly. Trying to reduce the full state
            # might cause infinite recursion loop. (Example: 2 new objects
            # reference each other.)
            doc = {}
            # Make sure that the object gets saved fully later.
            self._jar.register(obj)

            #doc_id = self._jar.createId()
            #oid = DBRef(table_name, doc_id, db_name)
            #return oid
        else:
            # XXX: Handle newargs; see ZODB.serialize.ObjectWriter.serialize
            # Go through each attribute and search for persistent references.
            doc = self.get_state(obj.__getstate__(), obj)

        # Always add a persistent type info
        doc[interfaces.PY_TYPE_ATTR_NAME] = get_dotted_name(obj.__class__)

        stored = False
        if interfaces.IColumnSerialization.providedBy(obj):
            self._jar._ensure_sql_columns(obj, table_name)
            column_data = obj._pj_get_column_fields()
        else:
            column_data = None
        if obj._p_oid is None:
            doc_id = self._jar._insert_doc(
                db_name, table_name, doc, id, column_data)
            stored = True
            obj._p_oid = DBRef(table_name, doc_id, db_name)
            # Make sure that any other code accessing this object in this
            # session, gets the same instance.
            self._jar._object_cache[hash(obj._p_oid)] = obj
        else:
            self._jar._update_doc(
                db_name, table_name, doc, obj._p_oid.id, column_data)
            stored = True
        # let's call the hook here, to always have _p_jar and _p_oid set
        if interfaces.IPersistentSerializationHooks.providedBy(obj):
            obj._pj_after_store_hook(self._jar._conn)

        if stored:
            # Make sure that the doc is added to the latest states.
            self._jar._latest_states[obj._p_oid] = doc

        return obj._p_oid


@zope.interface.implementer(interfaces.IObjectReader)
class ObjectReader(object):

    def __init__(self, jar):
        self._jar = jar
        self.preferPersistent = True

    def simple_resolve(self, path):
        path = path.replace('_dot_', '.')
        path = path[1:] if path.startswith('u_') else path
        # We try to look up the klass from a cache. The important part here is
        # that we also cache lookup failures as None, since they actually
        # happen more frequently than a hit due to an optimization in the
        # resolve() function.
        try:
            klass = PATH_RESOLVE_CACHE[path]
        except KeyError:
            if path in KNOWN_FACTORIES:
                return KNOWN_FACTORIES[path]
            try:
                klass = resolve(path)
            except ImportError:
                PATH_RESOLVE_CACHE[path] = klass = None
            else:
                PATH_RESOLVE_CACHE[path] = klass
        if klass is None:
            raise ImportError(path)
        return klass

    def resolve(self, dbref):
        __traceback_info__ = dbref
        # 1. Try to optimize on whether there's just one class stored in one
        #    table, that can save us one DB query
        if dbref.table in TABLE_KLASS_MAP:
            results = TABLE_KLASS_MAP[dbref.table]
            if len(results) == 1:
                # there must be just ONE, otherwise we need to check the JSONB
                klass = list(results)[0]
                return klass
        # from this point on we need the dbref.id
        if dbref.id is None:
            raise ImportError(dbref)
        # 2. Get the class from the object state
        #    Multiple object types are stored in the table. We have to
        #    look at the object (JSONB) to find out the type.
        if dbref in self._jar._latest_states:
            # Optimization: If we have the latest state, then we just get
            # this object document. This is used for fast loading or when
            # resolving the same object path a second time. (The latter
            # should never happen due to the object cache.)
            obj_doc = self._jar._latest_states[dbref]
        elif ALWAYS_READ_FULL_DOC:
            # Optimization: Read the entire doc and stick it in the right
            # place so that unghostifying the object later will not cause
            # another database access.
            obj_doc = self._jar._get_doc_by_dbref(dbref)
            # Do not pollute the latest states because the ref could not be
            # found.
            if obj_doc is not None:
                self._jar._latest_states[dbref] = obj_doc
        else:
            # Just read the type from the database, still requires one query
            pytype = self._jar._get_doc_py_type(
                dbref.database, dbref.table, dbref.id)
            obj_doc = {interfaces.PY_TYPE_ATTR_NAME: pytype}
        if obj_doc is None:
            # There is no document for this reference in the database.
            raise ImportError(dbref)
        if interfaces.PY_TYPE_ATTR_NAME in obj_doc:
            # We have always the path to the class in JSONB
            klass = self.simple_resolve(obj_doc[interfaces.PY_TYPE_ATTR_NAME])
        else:
            raise ImportError(dbref)
        return klass

    def _set_object_state(self, state, sub_obj, obj):
        sub_obj_state = self.get_object(state, obj)
        if hasattr(sub_obj, '__setstate__'):
            sub_obj.__setstate__(sub_obj_state)
        else:
            sub_obj.__dict__.update(sub_obj_state)
        if isinstance(sub_obj, persistent.Persistent):
            # This is a persistent sub-object -- mark it as such. Otherwise
            # we risk to store this object in its own table next time.
            setattr(sub_obj, interfaces.SUB_OBJECT_ATTR_NAME, True)

    def get_non_persistent_object(self, state, obj):
        if '_py_constant' in state:
            return self.simple_resolve(state['_py_constant'])

        # this method must NOT change the passed in state dict
        state = dict(state)
        if '_py_type' in state:
            # Handle the simplified case.
            klass = self.simple_resolve(state.pop('_py_type'))
            sub_obj = copyreg._reconstructor(klass, object, None)
            self._set_object_state(state, sub_obj, obj)
        elif interfaces.PY_TYPE_ATTR_NAME in state:
            # Another simple case for persistent objects that do not want
            # their own document.
            klass = self.simple_resolve(state.pop(interfaces.PY_TYPE_ATTR_NAME))
            sub_obj = copyreg.__newobj__(klass)
            self._set_object_state(state, sub_obj, obj)
        else:
            factory = self.simple_resolve(state.pop('_py_factory'))
            factory_args = self.get_object(state.pop('_py_factory_args'), obj)
            sub_obj = factory(*factory_args)
            # if there is anything left over in `state`, set it below
            # otherwise setting a {} state seems to clean out the object
            # but this is such an edge case of an edge case....
            if state:
                self._set_object_state(state, sub_obj, obj)

        if getattr(sub_obj, interfaces.SUB_OBJECT_ATTR_NAME, False):
            setattr(sub_obj, interfaces.DOC_OBJECT_ATTR_NAME, obj)
            sub_obj._p_jar = self._jar
        return sub_obj

    def get_object(self, state, obj):
        # stateIsDict and state_py_type: optimization to avoid X lookups
        # the code was:
        # if isinstance(state, dict) and state.get('_py_type') == 'DBREF':
        # this methods gets called a gazillion times, so being fast is crucial
        stateIsDict = isinstance(state, dict)
        if stateIsDict:
            state_py_type = state.get('_py_type')
            if state_py_type == 'BINARY':
                # Binary data in Python 2 is presented as a string. We will
                # convert back to binary when serializing again.
                return base64.b64decode(state['data'])
            if state_py_type == 'DBREF':
                # Load a persistent object. Using the _jar.load() method to make
                # sure we're loading from right database and caching is properly
                # applied.
                dbref = DBRef(state['table'], state['id'], state['database'])
                return self._jar.load(dbref)
            if state_py_type == 'type':
                # Convert a simple object reference, mostly classes.
                return self.simple_resolve(state['path'])
            if state_py_type == 'datetime.date':
                return datetime.datetime.strptime(
                    state['value'], FMT_DATE).date()
            if state_py_type == 'datetime.time':
                try:
                    return datetime.datetime.strptime(
                        state['value'], FMT_TIME).time()
                except ValueError:
                    # BBB: We originally did not track sub-seconds.
                    warnings.warn(
                        "Data in old time format found. Support for the "
                        "old format will be removed in pjpersist 2.0.",
                        DeprecationWarning)
                    return datetime.datetime.strptime(
                        state['value'], FMT_TIME_BBB).time()
            if state_py_type == 'datetime.datetime':
                try:
                    return datetime.datetime.strptime(
                        state['value'], FMT_DATETIME)
                except ValueError:
                    # BBB: We originally did not track sub-seconds.
                    warnings.warn(
                        "Data in old date/time format found. Support for the "
                        "old format will be removed in pjpersist 2.0.",
                        DeprecationWarning)
                    return datetime.datetime.strptime(
                        state['value'], FMT_DATETIME_BBB)

        # Give the custom serializers a chance to weigh in.
        for serializer in SERIALIZERS:
            if serializer.can_read(state):
                return serializer.read(state)

        if stateIsDict and (
            '_py_factory' in state
            or '_py_constant' in state
            or '_py_type' in state
            or interfaces.PY_TYPE_ATTR_NAME in state):
            # Load a non-persistent object.
            return self.get_non_persistent_object(state, obj)
        if isinstance(state, (tuple, list)):
            # All lists are converted to persistent lists, so that their state
            # changes are noticed. Also make sure that all value states are
            # converted to objects.
            sub_obj = [self.get_object(value, obj) for value in state]
            if self.preferPersistent:
                sub_obj = PersistentList(sub_obj)
                setattr(sub_obj, interfaces.DOC_OBJECT_ATTR_NAME, obj)
                sub_obj._p_jar = self._jar
            return sub_obj
        if stateIsDict:
            # All dictionaries are converted to persistent dictionaries, so
            # that state changes are detected. Also convert all value states
            # to objects.
            # Handle non-string key dicts.
            # Note: see comments at the definition of DICT_NON_STRING_KEY_MARKER
            if DICT_NON_STRING_KEY_MARKER in state:
                items = state[DICT_NON_STRING_KEY_MARKER]
                # a dict can NOT have an unhashable key,
                # most commonly used is a tuple (which converts to JSON as a
                # list, what then would fail here)
                # so convert a list-ish key to tuple
                keyConverter = lambda key: (
                    tuple(key)
                    if isinstance(key, (list, PersistentList))
                    else key)
                sub_obj = {
                    keyConverter(self.get_object(name, obj)):
                        self.get_object(value, obj)
                    for name, value in items}
            else:
                # for performance reasons we keep a separate implementation
                sub_obj = {
                    self.get_object(name, obj):
                        self.get_object(value, obj)
                    for name, value in state.items()}
            if self.preferPersistent:
                sub_obj = PersistentDict(sub_obj)
                setattr(sub_obj, interfaces.DOC_OBJECT_ATTR_NAME, obj)
                sub_obj._p_jar = self._jar
            return sub_obj
        return state

    def set_ghost_state(self, obj, doc=None):
        __traceback_info__ = (obj, doc)
        # Check whether the object state was stored on the object itself.
        if doc is None:
            doc = getattr(obj, interfaces.STATE_ATTR_NAME, None)
        # Look up the object state by table_name and oid.
        if doc is None:
            doc = self._jar._get_doc_by_dbref(obj._p_oid)
        # Check that we really have a state doc now.
        if doc is None:
            raise ImportError(obj._p_oid)
        # Remove unwanted attributes.
        doc.pop(interfaces.PY_TYPE_ATTR_NAME, None)
        # Now convert the document to a proper Python state dict.
        state = dict(self.get_object(doc, obj))
        if obj._p_oid not in self._jar._latest_states:
            # Sometimes this method is called to update the object state
            # before storage. Only update the latest states when the object is
            # originally loaded.
            self._jar._latest_states[obj._p_oid] = doc
        # Set the state.
        obj.__setstate__(state)
        # Run the custom load functions.
        if interfaces.IPersistentSerializationHooks.providedBy(obj):
            obj._pj_after_load_hook(self._jar._conn)

    def get_ghost(self, dbref, klass=None):
        # If we can, we return the object from cache.
        try:
            return self._jar._object_cache[hash(dbref)]
        except KeyError:
            pass
        if klass is None:
            klass = self.resolve(dbref)
        obj = klass.__new__(klass)
        obj._p_jar = self._jar
        obj._p_oid = dbref
        del obj._p_changed
        # Assign the table after deleting _p_changed, since the attribute
        # is otherwise deleted.
        setattr(obj, interfaces.DATABASE_ATTR_NAME, dbref.database)
        setattr(obj, interfaces.TABLE_ATTR_NAME, dbref.table)
        # Adding the object to the cache is very important, so that we get the
        # same object reference throughout the transaction.
        self._jar._object_cache[hash(dbref)] = obj
        return obj


class table(object):
    """Declare the table used by the class.

    sets also the atrtibute interfaces.TABLE_ATTR_NAME
    but register the fact also in TABLE_KLASS_MAP, this will allow pjpersist
    to optimize class lookup when just one class is stored in one table
    otherwise class lookup always needs the JSONB data from PG
    """

    def __init__(self, table_name):
        self.table_name = table_name

    def __call__(self, ob):
        try:
            setattr(ob, interfaces.TABLE_ATTR_NAME, self.table_name)
            TABLE_KLASS_MAP.setdefault(self.table_name, set()).add(ob)
        except AttributeError:
            raise TypeError(
                "Can't declare %s" % interfaces.TABLE_ATTR_NAME, ob)
        return ob



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Thread-aware PG/JSONB Connection Pool"""
import logging
import threading
import psycopg2
import zope.interface

from pjpersist import datamanager, interfaces

log = logging.getLogger('pjpersist')

LOCAL = threading.local()

# XXX: THIS SEEMS MAJORLY BROKEN< SINCE CONNECTIONS ARE NEVER RETURNED TO THE
# POOL.

@zope.interface.implementer(interfaces.IPJDataManagerProvider)
class PJDataManagerProvider(object):

    def __init__(self, user=None, password=None, host='localhost', port=5432,
                 pool_min_conn=1, pool_max_conn=8):
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.pools = {}

    def get(self, database):
        # Make sure the dict containing the local data managers exists.
        if not hatattr(LOCAL, 'dms'):
            LOCAL.dms = {}
        # Get the data manager, if it exists.
        try:
            return LOCAL.dms[database]
        except KeyError:
            pass
        # Create a new pool, if necessary.
        if database not in self.pools:
            self.pools[database] = psycopg2.pool.PersistentConnectionPool(
                database=database, user=self.user, password=self.password,
                host=self.host, port=self.port)
        # Create a new data manager and return it.
        LOCAL.dms[database] = datamanager.PJDataManager(
            self.pools[database].getconn())
        return LOCAL.dms[database]



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import unittest

from pjpersist import testing, sqlbuilder as sb


def run(expr):
    return expr.__sqlrepr__('postgres')


def doctest_sqlbuilder():
    r"""We have functions that correspond to the JSON operators:

        >>> run(sb.PGArray([sb.PGArray(['a', 'b']),
        ...         sb.PGArray(['c', 'd'])]))
        "array[array['a', 'b'], array['c', 'd']]"
        >>> run(sb.PGArray([sb.PGArray([1, 2]),
        ...         sb.PGArray([3, 4])]))
        'array[array[1, 2], array[3, 4]]'
        >>> run(sb.PGArrayLiteral(["a'b", 'b', 2, 3.5]))
        '\'{"a\'\'b", "b", 2, 3.5}\''
        >>> run(sb.PGArrayLiteral([sb.PGArrayLiteral(["a'b", 'b']),
        ...         sb.PGArrayLiteral(['c', 'd'])]))
        '\'{{"a\'\'b", "b"}, {"c", "d"}}\''
        >>> run(sb.Select(
        ...         sb.JSON_GETITEM(sb.table.stakeholder.data, "address")))
        "SELECT ((stakeholder.data) -> ('address')) FROM stakeholder"

        >>> run(sb.JSON_GETITEM_TEXT(sb.table.person.data, "address"))
        "((person.data) ->> ('address'))"

        >>> run(sb.JSON_PATH(sb.table.person.data,
        ...                  ["address", "zip"]))
        "((person.data) #> (array['address', 'zip']))"

        >>> run(sb.JSON_PATH_TEXT(sb.table.person.data,
        ...                       ["address", "zip"]))
        '((person.data) #>> (\'{"address", "zip"}\'))'

    We can cast JSON strings to the JSONB type:

        >>> print(run(sb.JSONB('{"universe": {"answer": 42}}')))
        '{"universe": {"answer": 42}}'::jsonb

    And perform JSONB "set operations" on the objects:

        >>> run(sb.JSONB_SUBSET(
        ...     sb.JSONB('{"address": {"zip": 22401}}'),
        ...     sb.table.Frob.data))
        '((\'{"address": {"zip": 22401}}\'::jsonb) <@ (Frob.data))'

        >>> run(sb.JSONB_SUPERSET(
        ...     sb.table.Frob.data,
        ...     sb.JSONB('{"address": {"zip": 22401}}')))
        '((Frob.data) @> (\'{"address": {"zip": 22401}}\'::jsonb))'

    Arrays work only on Postgres:

        >>> sb.JSON_PATH(sb.table.person.data,
        ...                  ["address", "zip"]).__sqlrepr__('sqlite')
        Traceback (most recent call last):
          ...
        AssertionError: Postgres-specific feature, sorry.

    Basic function support:

        >>> run(sb.Function('do', 1, 'foo'))
        "do(1, 'foo')"

        >>> run(sb.ARRAY_TO_STRING(sb.Field('table', 'col'), ', '))
        "array_to_string(table.col, ', ')"
    """


def doctest_sqlbuilder_with_queries():
    """
        >>> subsel = sb.WithSubquery(
        ...     'regional_sales',
        ...     sb.Select(['region', 'amount'], staticTables=['orders']),
        ...     ['r', 'a'])
        >>> run(subsel)
        'regional_sales (r, a) AS ( SELECT region, amount FROM orders )'

        >>> withstmt = sb.With(
        ...     [subsel],
        ...     sb.Select(['r'], staticTables=['regional_sales'])
        ... )

        >>> print(run(withstmt))
        WITH regional_sales (r, a) AS ( SELECT region, amount FROM orders )
        SELECT r FROM regional_sales

        >>> withstmt = sb.With(
        ...     [subsel],
        ...     sb.Select(['r'], staticTables=['regional_sales']),
        ...     recursive=True
        ... )

        >>> print(run(withstmt))
        WITH RECURSIVE regional_sales (r, a) AS (
            SELECT region, amount FROM orders )
        SELECT r FROM regional_sales

    """

def doctest_sqlbuilder_unions():
    """
        >>> union = sb.Union(
        ...     sb.Select(['r'], staticTables=['regional_sales']),
        ...     sb.Select(['r'], staticTables=['global_sales'])
        ... )
        >>> print(run(union))
        ( SELECT r FROM regional_sales )
        UNION
        ( SELECT r FROM global_sales )

        >>> union = sb.UnionAll(
        ...     sb.Select(['r'], staticTables=['regional_sales']),
        ...     sb.Select(['r'], staticTables=['global_sales'])
        ... )
        >>> print(run(union))
        ( SELECT r FROM regional_sales )
        UNION ALL
        ( SELECT r FROM global_sales )
    """


def test_suite():
    return unittest.TestSuite([
        doctest.DocTestSuite(
            optionflags=testing.OPTIONFLAGS),
        doctest.DocTestSuite(
            module=sb,
            optionflags=testing.OPTIONFLAGS),
        ])




##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Query statistics reporter tests"""

import doctest

from pjpersist import testing
from pjpersist.querystats import QueryReport


def doctest_calculate_empty():
    """
    Stats on fresh collector indicate 0 queries
        >>> qr = QueryReport()
        >>> qr.calc_totals()
        QueryTotals(total_queries=0, total_time=0, sorted_queries=[])
    """


def doctest_calculate_several():
    """
    Record several queries and calculate stats

        >>> qr = QueryReport()
        >>> qr.record("SELECT 1", [], 0.005)
        >>> qr.record("SELECT 2", [], 0.0001)
        >>> qr.record("SELECT 3", [], 0.8)

        >>> stats = qr.calc_totals()
        >>> stats
        QueryTotals(total_queries=3, total_time=0.8051, sorted_queries=[...])

        >>> [q.query for q in stats.sorted_queries]
        ['SELECT 2', 'SELECT 1', 'SELECT 3']
    """


def doctest_calc_and_report_empty():
    """
    Print out empty query report

        >>> qr = QueryReport()
        >>> print(qr.calc_and_report())
        Query report: no queries were executed
        """


def doctest_calc_and_report_several():
    """
    Record several queries and produce report

        >>> qr = QueryReport()
        >>> qr.record("SELECT 1", [], 0.005)
        >>> qr.record("SELECT 2", (1, 2, 3), 0.0001)
        >>> qr.record("SELECT 3", ["a", "b", 3], 0.8)
        >>> qr.report_traceback = True

        >>> print(qr.calc_and_report())
        Query report:
        ------------------------------------------------------------
        10 most expensive queries:
        *** SELECT 2
        ... ARGS: (1, 2, 3)
        ... TIME: 0.1000ms
        File ...
        <BLANKLINE>
        *** SELECT 1
        ... ARGS: []
        ... TIME: 5.0000ms
        File ...
        <BLANKLINE>
        *** SELECT 3
        ... ARGS: ['a', 'b', 3]
        ... TIME: 800.0000ms
        File ...
        <BLANKLINE>
        ------------------------------------------------------------
        Queries executed: 3
        Time spent: 805.1000ms
    """


def test_suite():
    dtsuite = doctest.DocTestSuite(
        optionflags=testing.OPTIONFLAGS)

    return dtsuite




##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


def patch_ujson_into_psycopg2():
    # Use fast json parser.
    try:
        import ujson
    except ImportError:
        return

    # patch decoding
    from psycopg2.extras import register_default_json, register_default_jsonb
    register_default_json(globally=True, loads=ujson.loads)
    register_default_jsonb(globally=True, loads=ujson.loads)

    # patch encoding
    from psycopg2 import _json
    _json.json = ujson



"""Setup
"""
import os
from setuptools import setup, find_packages


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames), 'rb') as f:
        return f.read().decode('utf-8')


def alltests():
    import os
    import sys
    import unittest
    # use the zope.testrunner machinery to find all the
    # test suites we've put under ourselves
    import zope.testrunner.find
    import zope.testrunner.options
    here = os.path.abspath(os.path.join(os.path.dirname(__file__), 'src'))
    args = sys.argv[:]
    defaults = ["--test-path", here]
    options = zope.testrunner.options.get_options(args, defaults)
    suites = list(zope.testrunner.find.find_suites(options))
    return unittest.TestSuite(suites)


TESTS_REQUIRE = [
    'zope.testrunner',
    'zope.app.testing',
    'zope.testing',
    'ZODB',
    'mock',
   ]


setup(
    name='pjpersist',
    version='3.0.2',
    author="Shoobx Team",
    author_email="dev@shoobx.com",
    url='https://github.com/Shoobx/pjpersist',
    description="PostgreSQL/JSONB Persistence Backend",
    long_description=(
        read('src', 'pjpersist', 'README.txt')
        + '\n\n' +
        read('CHANGES.rst')
    ),
    license="ZPL 2.1",
    keywords="postgres jsonb persistent",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: ZODB',
        'License :: OSI Approved :: Zope Public License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    data_files=[('pjpersist', ['README.rst'])],  # in order to test it with tox
    extras_require = dict(
        test=TESTS_REQUIRE,
        zope=(
            'zope.container',
        ),
    ),
    install_requires=[
        'future',
        'persistent',
        'transaction',
        'repoze.lru',
        'psycopg2',
        'simplejson',
        'setuptools',
        'sqlobject',
        'zope.dottedname',
        'zope.interface',
        'zope.schema',
        'zope.exceptions >=3.7.1',  # required for extract_stack
    ],
    include_package_data=True,
    zip_safe=False,
    entry_points='''
    [console_scripts]
    profile = pjpersist.tests.performance:main
    json_speed_test = pjpersist.tests.json_speed_test:main
    ''',
    tests_require=TESTS_REQUIRE,
    test_suite='__main__.alltests',
)



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostGreSQL/JSONB Persistence Serialization Tests"""
import base64
import collections
import datetime
import doctest
import persistent
import pprint
import copy
import copyreg
import pickle

from pjpersist import interfaces, serialize, testing

class Top(persistent.Persistent):
    _p_pj_table = 'Top'

def create_top(name):
    top = Top()
    top.name = name
    return top

class Top2(Top):
    pass

class Tier2(persistent.Persistent):
    _p_pj_sub_object = True

class Foo(persistent.Persistent):
    _p_pj_table = 'Foo'

class Bar(persistent.Persistent):
    _p_pj_database = 'foo'
    _p_pj_table = 'Bar'

class Anything(persistent.Persistent):
    pass

class StoreType(persistent.Persistent):
    _p_pj_table = 'storetype'

class StoreType2(StoreType):
    pass

class Simple(object):
    pass

class Constant(object):
    def __reduce__(self):
        return 'Constant'
Constant = Constant()

class StringConstant(str):
    def __reduce__(self):
        return 'StringConstant'
StringConstant = StringConstant()

class CopyReggedConstant(object):
    def custom_reduce_fn(self):
        return 'CopyReggedConstant'
copyreg.pickle(CopyReggedConstant, CopyReggedConstant.custom_reduce_fn)
CopyReggedConstant = CopyReggedConstant()


def doctest_link_to_parent():
    """Link to Parent

    When an object is written, all persistent subobjects will receive a jar
    and a link to the parent object. However, if one develops their own
    `__getstate__` and `__setstate__` functions, this assignment must be done
    manually. This method provides a quick API for doing so.

    >>> foo = Foo()
    >>> commit()

    >>> foo.bar = Bar('bar')
    >>> foo.bar._p_jar is None
    True

    >>> serialize.link_to_parent(foo.bar, foo)
    >>> foo.bar._p_jar is foo._p_jar
    True
    >>> foo.bar._p_pj_doc_object is foo
    True
    """


def doctest_link_to_parent_full_example():
    """Link to Parent - Full Example

    >>> class Blah(persistent.Persistent):
    ...     def __init__(self):
    ...         self.dict = serialize.PersistentDict()
    ...     def __getstate__(self):
    ...         serialize.link_to_parent(self.dict, self)
    ...         return {'dict': dict(self.dict)}
    ...     def __setstate__(self, state):
    ...         self.dict = serialize.PersistentDict(state['dict'])

    >>> blah = Blah()
    >>> blah.dict._p_jar is None
    True

    >>> dm.root['blah'] = blah

    >>> blah.dict._p_jar is dm
    True

    >>> blah.dict[1] = 'data'
    >>> blah.dict._p_changed
    True

    >>> commit()
    >>> dm.root['blah'].dict
    {1: 'data'}
    """


def doctest_DBRef():
    """DBRef class

    Create a simple DBRef to start with:

      >>> dbref1 = serialize.DBRef('table1', '0001', 'database1')
      >>> dbref1
      DBRef('table1', '0001', 'database1')

    We can also convert the ref quickly to a JSON structure or a simple tuple:

      >>> dbref1.as_tuple()
      ('database1', 'table1', '0001')

      >>> pprint.pprint(dbref1.as_json())
      {'_py_type': 'DBREF',
       'database': 'database1',
       'id': '0001',
       'table': 'table1'}

    Note that the hash of a ref is consistent over all DBRef instances:

      >>> dbref11 = serialize.DBRef('table1', '0001', 'database1')
      >>> hash(dbref1) == hash(dbref11)
      True

    Let's make sure that some other comparisons work as well:

      >>> dbref1 == dbref11
      True

      >>> dbref1 in [dbref11]
      True

    Let's now compare to a truely different DB Ref instance:

      >>> dbref2 = serialize.DBRef('table1', '0002', 'database1')

      >>> hash(dbref1) == hash(dbref2)
      False
      >>> dbref1 == dbref2
      False
      >>> dbref1 in [dbref2]
      False

    Serialization also works well.

      >>> refp = pickle.dumps(dbref1)

      >>> dbref11 = pickle.loads(refp)
      >>> dbref1 == dbref11
      True
      >>> id(dbref1) == id(dbref11)
      False

    Don't fail on various comparisons:

      >>> dbref1 == None
      False

      >>> dbref1 == Simple
      False

      >>> dbref1 != None
      True

      >>> dbref1 != Simple
      True

    """

def doctest_ObjectSerializer():
    """Test the abstract ObjectSerializer class.

    Object serializers are hooks into the serialization process to allow
    better serialization for particular objects. For example, the result of
    reducing a datetime.date object is a short, optimized binary string. This
    representation might be optimal for pickles, but is really aweful for
    PostGreSQL, since it does not allow querying for dates. An object
    serializer can be used to use a better representation, such as the date
    ordinal number.

      >>> os = serialize.ObjectSerializer()

    So here are the methods that must be implemented by an object serializer:

      >>> os.can_read({})
      Traceback (most recent call last):
      ...
      NotImplementedError

      >>> os.read({})
      Traceback (most recent call last):
      ...
      NotImplementedError

      >>> os.can_write(object())
      Traceback (most recent call last):
      ...
      NotImplementedError

      >>> os.write(object())
      Traceback (most recent call last):
      ...
      NotImplementedError
    """

def doctest_ObjectWriter_get_table_name():
    """ObjectWriter: get_table_name()

    This method determines the table name and database for a given
    object. It can either be specified via '_p_pj_table' or is
    determined from the class path. When the table name is specified, the
    mapping from table name to class path is stored.

      >>> writer = serialize.ObjectWriter(dm)
      >>> writer.get_table_name(Anything())
      ('pjpersist_test', 'pjpersist_dot_tests_dot_test_serialize_dot_Anything')

      >>> top = Top()
      >>> writer.get_table_name(top)
      ('pjpersist_test', 'Top')

    When classes use inheritance, it often happens that all sub-objects share
    the same table. However, only one can have an entry in our mapping
    table to avoid non-unique answers. Thus we require all sub-types after the
    first one to store their typing providing a hint for deseriealization:

      >>> top2 = Top2()
      >>> writer.get_table_name(top2)
      ('pjpersist_test', 'Top')

    Since the serializer also supports serializing any object without the
    intend of storing it in PostGreSQL, we have to be abel to look up the
    table name of a persistent object without a jar being around.

      >>> writer = serialize.ObjectWriter(None)
      >>> writer.get_table_name(Bar())
      ('foo', 'Bar')

    """

def doctest_ObjectWriter_get_non_persistent_state():
    r"""ObjectWriter: get_non_persistent_state()

    This method produces a proper reduced state for custom, non-persistent
    objects.

      >>> writer = serialize.ObjectWriter(dm)

    A simple new-style class:

      >>> class This(object):
      ...     def __init__(self, num):
      ...         self.num = num

      >>> this = This(1)
      >>> pprint.pprint(writer.get_non_persistent_state(this))
      {'_py_type': '__main__.This', 'num': 1}

    A simple old-style class:

      >>> class That(object):
      ...     def __init__(self, num):
      ...         self.num = num

      >>> that = That(1)
      >>> pprint.pprint(writer.get_non_persistent_state(that))
      {'_py_type': '__main__.That', 'num': 1}

    The method also handles persistent classes that do not want their own
    document:

      >>> top = Top()
      >>> writer.get_non_persistent_state(top)
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top'}

    And then there are the really weird cases, which is the reason we usually
    have serializers for them:

      >>> orig_serializers = serialize.SERIALIZERS
      >>> serialize.SERIALIZERS = []

      >>> pprint.pprint(
      ...     writer.get_non_persistent_state(datetime.date(2011, 11, 1)))
      {'_py_factory': 'datetime.date',
       '_py_factory_args': [{'_py_type': 'BINARY', 'data': 'B9sLAQ=='}]}

      >>> serialize.SERIALIZERS = orig_serializers
    """

def doctest_ObjectWriter_get_persistent_state():
    r"""ObjectWriter: get_persistent_state()

    This method produces a proper reduced state for a persistent object, which
    is basically a DBRef.

      >>> writer = serialize.ObjectWriter(dm)

      >>> foo = Foo()
      >>> foo._p_oid

      >>> pprint.pprint(writer.get_persistent_state(foo))
      {'_py_type': 'DBREF',
       'database': 'pjpersist_test',
       'id': '0001020304050607080a0b0c0',
       'table': 'Foo'}

      >>> dm.commit(None)
      >>> foo._p_oid
      DBRef('Foo', '0001020304050607080a0b0c0', 'pjpersist_test')
      >>> dumpTable('Foo')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Foo'},
        'id': '0001020304050607080a0b0c0'}]

    The next time the object simply returns its reference:

      >>> pprint.pprint(writer.get_persistent_state(foo))
      {'_py_type': 'DBREF',
       'database': 'pjpersist_test',
       'id': '0001020304050607080a0b0c0',
       'table': 'Foo'}
      >>> dumpTable('Foo')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Foo'},
        'id': '0001020304050607080a0b0c0'}]
    """


def doctest_ObjectWriter_get_state_PJ_NATIVE_TYPES():
    """ObjectWriter: get_state(): PJ-native Types

      >>> writer = serialize.ObjectWriter(None)
      >>> writer.get_state(1)
      1
      >>> writer.get_state(1.0)
      1.0
      >>> writer.get_state('Test')
      'Test'
      >>> print(writer.get_state(None))
      None
    """

def doctest_ObjectWriter_get_state_constant():
    """ObjectWriter: get_state(): Constants

      >>> writer = serialize.ObjectWriter(None)
      >>> writer.get_state(Constant)
      {'_py_constant': 'pjpersist.tests.test_serialize.Constant'}
      >>> writer.get_state(interfaces.IObjectWriter)
      {'_py_constant': 'pjpersist.interfaces.IObjectWriter'}
      >>> writer.get_state(CopyReggedConstant)
      {'_py_constant': 'pjpersist.tests.test_serialize.CopyReggedConstant'}
      >>> writer.get_state(StringConstant)
      {'_py_constant': 'pjpersist.tests.test_serialize.StringConstant'}
    """

def doctest_ObjectWriter_get_state_types():
    """ObjectWriter: get_state(): types (type, class)

      >>> writer = serialize.ObjectWriter(None)
      >>> pprint.pprint(writer.get_state(Top))
      {'_py_type': 'type', 'path': 'pjpersist.tests.test_serialize.Top'}
      >>> pprint.pprint(writer.get_state(str))
      {'_py_type': 'type', 'path': '__builtin__.str'}
    """

def doctest_ObjectWriter_get_state_sequences():
    """ObjectWriter: get_state(): sequences (tuple, list, PersistentList)

    We convert any sequence into a simple list, since JSONB supports that
    type natively. But also reduce any sub-objects.

      >>> class Number(object):
      ...     def __init__(self, num):
      ...         self.num = num

      >>> writer = serialize.ObjectWriter(None)
      >>> pprint.pprint(writer.get_state((1, '2', Number(3))))
      [1, '2', {'_py_type': '__main__.Number', 'num': 3}]
      >>> pprint.pprint(writer.get_state([1, '2', Number(3)]))
      [1, '2', {'_py_type': '__main__.Number', 'num': 3}]
    """

def doctest_ObjectWriter_get_state_mappings():
    """ObjectWriter: get_state(): mappings (dict, PersistentDict)

    We convert any mapping into a simple dict, since JSONB supports that
    type natively. But also reduce any sub-objects.

      >>> class Number(object):
      ...     def __init__(self, num):
      ...         self.num = num

      >>> writer = serialize.ObjectWriter(None)
      >>> pprint.pprint(writer.get_state({'1': 1, '2': '2', '3': Number(3)}))
      {'1': 1, '2': '2', '3': {'_py_type': '__main__.Number', 'num': 3}}

    Unfortunately, JSONB only supports text keys. So whenever we have non-text
    keys, we need to create a less natural, but consistent structure:

      >>> pprint.pprint(writer.get_state({1: 'one', 2: 'two', 3: 'three'}))
      {'dict_data': [(1, 'one'), (2, 'two'), (3, 'three')]}

    Another edge case is when the input has a `dict_data` key

      >>> writer = serialize.ObjectWriter(None)
      >>> state = writer.get_state(
      ...     {'1': 1, 'dict_data': 'works?'})
      >>> pprint.pprint(sorted(state['dict_data']))
      [('1', 1), ('dict_data', 'works?')]

    It gets even worse with tuple keyed mappings,
    tuple keys get converted to JSON as a list,
    which need to be handled then by get_object

      >>> mapping = {
      ...     (1, 'key-one', None): 'value-one',
      ...     (2, 'key-two', True): 'value-two',
      ...     'key-three': 'value-three'}
      >>> pprint.pprint(writer.get_state(mapping))
      {'dict_data': [([1, 'key-one', None], 'value-one'),
                     ([2, 'key-two', True], 'value-two'),
                     ('key-three', 'value-three')]}

    """

def doctest_ObjectWriter_get_state_Persistent():
    """ObjectWriter: get_state(): Persistent objects

      >>> writer = serialize.ObjectWriter(dm)

      >>> top = Top()
      >>> pprint.pprint(writer.get_state(top))
      {'_py_type': 'DBREF',
       'database': 'pjpersist_test',
       'id': '0001020304050607080a0b0c',
       'table': 'Top'}

    But a persistent object can declare that it does not want a separate
    document:

      >>> top2 = Top()
      >>> top2._p_pj_sub_object = True
      >>> writer.get_state(top2, top)
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top'}
    """

def doctest_ObjectWriter_get_state_sub_doc_object_with_no_pobj():
    """ObjectWriter: get_state(): Called with a sub-document object and no pobj

    While this should not really happen, we want to make sure we are properly
    protected against it. Usually, the writer sets the jar of the parent
    object equal to its jar. But it cannot do so, if `pobj` or `pobj._p_jar`
    is `None`.

      >>> writer = serialize.ObjectWriter(dm)

      >>> t2 = Tier2()
      >>> writer.get_state(t2)
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Tier2'}

      >>> t2._p_jar is None
      True
      >>> t2._p_pj_doc_object is None
      True

    Let's now pass in a `pobj` without a jar:

      >>> top = Top()
      >>> writer.get_state(t2, top)
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Tier2'}

      >>> t2._p_jar is None
      True
      >>> t2._p_pj_doc_object is top
      True
    """

def doctest_ObjectWriter_get_state_same_obj_in_dict():
    """ObjectWriter: get_state():

    If get_state gets the very same object in a structure it must not fail.

      >>> writer = serialize.ObjectWriter(dm)

      >>> pdict = serialize.PersistentDict()
      >>> one = Simple()
      >>> one.data = 'data'
      >>> pdict['one'] = one
      >>> pdict['two'] = one

      >> from cPickle import dumps

      >> print dumps(pdict)

      >> import pickletools
      >> pickletools.dis(dumps(pdict))

      >>> pprint.pprint(writer.get_state(pdict))
      {'one': {'_py_type': 'pjpersist.tests.test_serialize.Simple',
               'data': 'data'},
       'two': {'_py_type': 'pjpersist.tests.test_serialize.Simple',
               'data': 'data'}}

    """

def doctest_ObjectWriter_get_state_same_obj_in_list():
    """ObjectWriter: get_state():

    If get_state gets the very same object in a structure it must not fail.

      >>> writer = serialize.ObjectWriter(dm)

      >>> plist = serialize.PersistentList()
      >>> one = Simple()
      >>> one.data = 'data'
      >>> plist.append(one)
      >>> plist.append(one)

      >> from cPickle import dumps

      >> print dumps(plist, protocol=0)

      >>> pprint.pprint(writer.get_state(plist))
      [{'_py_type': 'pjpersist.tests.test_serialize.Simple', 'data': 'data'},
       {'_py_type': 'pjpersist.tests.test_serialize.Simple', 'data': 'data'}]

    """

def doctest_ObjectWriter_get_full_state():
    """ObjectWriter: get_full_state()

      >>> writer = serialize.ObjectWriter(dm)

    Let's get the state of a regular object"

      >>> any = Anything()
      >>> any.name = 'anything'
      >>> pprint.pprint(writer.get_full_state(any))
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Anything',
       'name': 'anything'}

      >>> any_ref = dm.insert(any)
      >>> pprint.pprint(writer.get_full_state(any))
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Anything',
       'name': 'anything'}

    Now an object that stores its type:

      >>> st = StoreType()
      >>> st.name = 'storetype'
      >>> pprint.pprint(writer.get_full_state(st))
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.StoreType',
       'name': 'storetype'}

      >>> st_ref = dm.insert(st)
      >>> pprint.pprint(writer.get_full_state(st))
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.StoreType',
       'name': 'storetype'}
    """

def doctest_ObjectWriter_store():
    """ObjectWriter: store()

      >>> writer = serialize.ObjectWriter(dm)

    Simply store an object:

      >>> top = Top()
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')
      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top'},
        'id': '0001020304050607080a0b0c0'}]

    Now that we have an object, storing an object simply means updating the
    existing document:

      >>> top.name = 'top'
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')
      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'name': 'top'},
        'id': '0001020304050607080a0b0c0'}]
    """

def doctest_ObjectWriter_store_with_new_object_references():
    """ObjectWriter: store(): new object references

    When two new objects reference each other, extracting the full state would
    cause infinite recursion errors. The code protects against that by
    optionally only creating an initial empty reference document.

      >>> writer = serialize.ObjectWriter(dm)

      >>> top = Top()
      >>> top.foo = Foo()
      >>> top.foo.top = top
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')
      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'foo': {'_py_type': 'DBREF',
                          'database': 'pjpersist_test',
                          'id': '0001020304050607080a0b0c0',
                          'table': 'Foo'}},
        'id': '0001020304050607080a0b0c0'}]
    """

def doctest_ObjectWriter_store_sub_persistent():
    """ObjectWriter: store()
    Make sure that subobject modifications get noticed after the
    first object add

      >>> writer = serialize.ObjectWriter(dm)

    Simply store an object:

      >>> top = Top()
      >>> top.top = Tier2()
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')
      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'top': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Tier2'}},
        'id': '0001020304050607080a0b0c0'}]

    Now that we have an object, update the subobject property:

      >>> top.top.name = 'top'

    JUST commit here to see whether the subobject changes are noticed by the
    persistence system

      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'top': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Tier2',
                          'name': 'top'}},
        'id': '0001020304050607080a0b0c0'}]
    """

def doctest_ObjectWriter_store_notsub_persistent():
    """ObjectWriter: store()

      >>> writer = serialize.ObjectWriter(dm)

    Simply store an object:

      >>> top = Top()
      >>> top.top = Foo()
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')
      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'top': {'_py_type': 'DBREF',
                          'database': 'pjpersist_test',
                          'id': '0001020304050607080a0b0c0',
                          'table': 'Foo'}},
        'id': '0001020304050607080a0b0c0'}]
      >>> dumpTable('Foo')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Foo'},
        'id': '5790ba2db25d2b42d000ccd3'}]

    Now that we have an object, update the subobject property:

      >>> top.top.name = 'top'

    JUST commit here to see whether the subobject changes are noticed by the
    persistence system

      >>> dm.commit(None)
      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'top': {'_py_type': 'DBREF',
                          'database': 'pjpersist_test',
                          'id': '0001020304050607080a0b0c0',
                          'table': 'Foo'}},
        'id': '0001020304050607080a0b0c0'}]

    Here we have the name set:

      >>> dumpTable('Foo')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Foo',
                 'name': 'top'},
        'id': '5790ba49b25d2b494600d22d'}]
    """

def doctest_ObjectReader_simple_resolve():
    """ObjectReader: simple_resolve()

    This methods simply resolves a Python path to the represented object.

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.simple_resolve('pjpersist.tests.test_serialize.Top')
      <class 'pjpersist.tests.test_serialize.Top'>

    After the original lookup, the result is cached:

      >>> pprint.pprint(serialize.PATH_RESOLVE_CACHE)
      {'pjpersist.tests.test_serialize.Top':
          <class 'pjpersist.tests.test_serialize.Top'>}

    Note that even lookup failures are cached.

      >>> reader.simple_resolve('path.to.bad')
      Traceback (most recent call last):
      ...
      ImportError: path.to.bad

      >>> pprint.pprint(serialize.PATH_RESOLVE_CACHE)
      {'path.to.bad': None,
       'pjpersist.tests.test_serialize.Top': <class 'pjpersist...Top'>}

     Resolving the path the second time uses the cache:

      >>> reader.simple_resolve('pjpersist.tests.test_serialize.Top')
      <class 'pjpersist.tests.test_serialize.Top'>

      >>> reader.simple_resolve('path.to.bad')
      Traceback (most recent call last):
      ...
      ImportError: path.to.bad
    """

def doctest_ObjectReader_resolve_simple_dblookup():
    """ObjectReader: resolve(): simple

    This methods resolves a table name to its class. The table name
    can be either any arbitrary string or a Python path.

      >>> reader = serialize.ObjectReader(dm)
      >>> ref = serialize.DBRef('Top', '4eb1b3d337a08e2de7000100')

    Now we need the doc to exist in the DB to be able to tell it's class.

      >>> reader.resolve(ref)
      Traceback (most recent call last):
      ...
      ImportError: DBRef('Top', '4eb1b3d337a08e2de7000100', None)
    """

def doctest_ObjectReader_resolve_simple_decorator():
    """ObjectReader: resolve(): decorator declared table

    This methods resolves a table name to its class. The table name
    can be either any arbitrary string or a Python path.

      >>> @serialize.table('foobar_table')
      ... class Foo(object):
      ...     pass

      >>> reader = serialize.ObjectReader(dm)
      >>> ref = serialize.DBRef('foobar_table', '4eb1b3d337a08e2de7000100')

    Once we declared on the class which table it uses, it's easy to resolve
    even without DB access.

      >>> result = reader.resolve(ref)
      >>> result
      <class '__main__.Foo'>

      >>> result is Foo
      True
    """

def doctest_ObjectReader_resolve_simple_decorator_more():
    """ObjectReader: resolve():
    decorator declared table, more classes in one table

    This methods resolves a table name to its class. The table name
    can be either any arbitrary string or a Python path.

      >>> @serialize.table('foobar_table')
      ... class FooBase(object):
      ...     pass

      >>> @serialize.table('foobar_table')
      ... class FooFoo(FooBase):
      ...     pass

      >>> reader = serialize.ObjectReader(dm)
      >>> ref = serialize.DBRef('foobar_table', '4eb1b3d337a08e2de7000100')

    As we have now more classes declared for the same table, we have to
    lookup the JSONB from the DB

      >>> result = reader.resolve(ref)
      Traceback (most recent call last):
      ...
      ImportError: DBRef('foobar_table', '4eb1b3d337a08e2de7000100', None)
    """

def doctest_ObjectReader_resolve_quick_when_type_in_doc():
    """ObjectReader: resolve(): Quick lookup when type in document.

    This methods resolves a table name to its class. The table name
    can be either any arbitrary string or a Python path.

      >>> st = StoreType()
      >>> st_ref = dm.insert(st)
      >>> st2 = StoreType2()
      >>> st2_ref = dm.insert(st2)
      >>> dm.commit(None)

    Let's now resolve the references:

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.resolve(st_ref)
      <class 'pjpersist.tests.test_serialize.StoreType'>
      >>> reader.resolve(st2_ref)
      <class 'pjpersist.tests.test_serialize.StoreType2'>

      >>> dm.commit(None)

    So here comes the trick. When fast-loading objects, the documents are made
    immediately available in the ``_latest_states`` mapping. This allows our
    quick resolve to utilize that document instead of looking it up in the
    database:

      >>> writer = serialize.ObjectWriter(dm)
      >>> tbl = dm._get_table_from_object(st)
      >>> dm._latest_states[st_ref] = writer.get_full_state(st)
      >>> dm._latest_states[st2_ref] = writer.get_full_state(st2)

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.resolve(st_ref)
      <class 'pjpersist.tests.test_serialize.StoreType'>
      >>> reader.resolve(st2_ref)
      <class 'pjpersist.tests.test_serialize.StoreType2'>

  """

def doctest_ObjectReader_resolve_lookup_with_multiple_maps():
    """ObjectReader: resolve(): lookup with multiple maps entries

    When the table name to Python path map has multiple entries, things
    are more interesting. In this case, we need to lookup the object, if it
    stores its persistent type otherwise we use the first map entry.

      >>> writer = serialize.ObjectWriter(dm)
      >>> top = Top()
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')
      >>> top2 = Top2()
      >>> writer.store(top2)
      DBRef('Top', '000000000000000000000001', 'pjpersist_test')
      >>> dm.commit(None)

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.resolve(top._p_oid)
      <class 'pjpersist.tests.test_serialize.Top'>
      >>> reader.resolve(top2._p_oid)
      <class 'pjpersist.tests.test_serialize.Top2'>

      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top'},
        'id': '0001020304050607080a0b0c0'},
       {'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top2'},
        'id': '0001020304050607080a0b0c0'}]

    If the DBRef does not have an object id, then an import error is raised:

      >>> reader.resolve(serialize.DBRef('Top', None, 'pjpersist_test'))
      Traceback (most recent call last):
      ...
      ImportError: DBRef('Top', None, 'pjpersist_test')
    """

def doctest_ObjectReader_get_non_persistent_object_py_type():
    """ObjectReader: get_non_persistent_object(): _py_type

    The simplest case is a document with a _py_type:

      >>> reader = serialize.ObjectReader(dm)
      >>> state = {'_py_type': 'pjpersist.tests.test_serialize.Simple'}
      >>> save_state = copy.deepcopy(state)
      >>> reader.get_non_persistent_object(state, None).__class__
      <class 'pjpersist.tests.test_serialize.Simple'>

    Make sure that state is unchanged:

      >>> state == save_state
      True

    It is a little bit more interesting when there is some additional state:

      >>> state = {'_py_type': 'pjpersist.tests.test_serialize.Simple',
      ...          'name': 'Here'}
      >>> save_state = copy.deepcopy(state)

      >>> simple = reader.get_non_persistent_object(state, None)
      >>> simple.name
      'Here'

    Make sure that state is unchanged:

      >>> state == save_state
      True

    """

def doctest_ObjectReader_get_non_persistent_object_py_persistent_type():
    """ObjectReader: get_non_persistent_object(): _py_persistent_type

    In this case the document has a _py_persistent_type attribute, which
    signals a persistent object living in its parent's document:

      >>> top = Top()

      >>> reader = serialize.ObjectReader(dm)
      >>> state = {'_py_persistent_type': 'pjpersist.tests.test_serialize.Tier2',
      ...          'name': 'Number 2'}
      >>> save_state = copy.deepcopy(state)

      >>> tier2 = reader.get_non_persistent_object(state, top)
      >>> tier2.__class__
      <class 'pjpersist.tests.test_serialize.Tier2'>

    We keep track of the containing object, so we can set _p_changed when this
    object changes.

      >>> tier2._p_pj_doc_object.__class__
      <class 'pjpersist.tests.test_serialize.Top'>
      >>> tier2._p_jar.__class__
      <class 'pjpersist.datamanager.PJDataManager'>

    Make sure that state is unchanged:

      >>> state == save_state
      True

    """

def doctest_ObjectReader_get_non_persistent_object_py_factory():
    """ObjectReader: get_non_persistent_object(): _py_factory

    This is the case of last resort. Specify a factory and its arguments:

      >>> reader = serialize.ObjectReader(dm)

      >>> state = {'_py_factory': 'pjpersist.tests.test_serialize.create_top',
      ...          '_py_factory_args': ('TOP',)}
      >>> save_state = copy.deepcopy(state)

      >>> top = reader.get_non_persistent_object(state, None)
      >>> top.__class__
      <class 'pjpersist.tests.test_serialize.Top'>
      >>> top.name
      'TOP'

    Make sure that state is unchanged:

      >>> state == save_state
      True

    """

def doctest_ObjectReader_get_object_binary():
    """ObjectReader: get_object(): binary data

    Binary data is just converted to bytes:

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object(
      ...     {
      ...         '_py_type': 'BINARY',
      ...         'data': base64.b64encode(b'hello'),
      ...     },
      ...     None) == b'hello'
      True
    """

def doctest_ObjectReader_get_object_datetime():
    """ObjectReader: get_object(): datetime

    Serialization of date/time related objects has a special handler.

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object(
      ...     {
      ...         '_py_type': 'datetime.date',
      ...         'value': '2005-07-13',
      ...     },
      ...     None)
      datetime.date(2005, 7, 13)
      >>> reader.get_object(
      ...     {
      ...         '_py_type': 'datetime.time',
      ...         'value': '17:18:10.100000',
      ...     },
      ...     None)
      datetime.time(17, 18, 10, 100000)
      >>> reader.get_object(
      ...     {
      ...         '_py_type': 'datetime.datetime',
      ...         'value': '2005-07-13T17:18:10.100000',
      ...     },
      ...     None)
      datetime.datetime(2005, 07, 13, 17, 18, 10, 100000)
    """

def doctest_ObjectReader_get_object_datetime_BBB():
    """ObjectReader: get_object(): datetime (BBB)

    Originally, we did not track sub-seconds.

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object(
      ...     {
      ...         '_py_type': 'datetime.time',
      ...         'value': '17:18:10',
      ...     },
      ...     None)
      datetime.time(17, 18, 10)
      >>> reader.get_object(
      ...     {
      ...         '_py_type': 'datetime.datetime',
      ...         'value': '2005-07-13T17:18:10',
      ...     },
      ...     None)
      datetime.datetime(2005, 7, 13, 17, 18, 10)
    """

def doctest_ObjectReader_get_object_dbref():
    """ObjectReader: get_object(): DBRef

      >>> writer = serialize.ObjectWriter(dm)
      >>> top = Top()
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')

    Database references load the ghost state of the object they represent:

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object(top._p_oid.as_json(), None).__class__
      <class 'pjpersist.tests.test_serialize.Top'>
    """

def doctest_ObjectReader_get_object_type_ref():
    """ObjectReader: get_object(): type reference

    Type references are resolved.

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object(
      ...     {'_py_type': 'type',
      ...      'path': 'pjpersist.tests.test_serialize.Simple'},
      ...     None)
      <class 'pjpersist.tests.test_serialize.Simple'>
    """

def doctest_ObjectReader_get_object_instance():
    """ObjectReader: get_object(): instance

    Instances are completely loaded:

      >>> reader = serialize.ObjectReader(dm)
      >>> simple = reader.get_object(
      ...     {'_py_type': 'pjpersist.tests.test_serialize.Simple',
      ...      'name': 'easy'},
      ...     None)
      >>> simple.__class__
      <class 'pjpersist.tests.test_serialize.Simple'>
      >>> simple.name
      'easy'
    """

def doctest_ObjectReader_get_object_sequence():
    """ObjectReader: get_object(): sequence

    Sequences become persistent lists with all objects deserialized.

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object([1, '2', 3.0], None)
      [1, '2', 3.0]
    """

def doctest_ObjectReader_get_object_mapping():
    """ObjectReader: get_object(): mapping

    Mappings become persistent dicts with all objects deserialized.

      >>> reader = serialize.ObjectReader(dm)
      >>> pprint.pprint(dict(reader.get_object({'1': 1, '2': 2, '3': 3}, None)))
      {'1': 1, '2': 2, '3': 3}

    Since JSONB does not allow for non-string keys, the state for a dict with
    non-string keys looks different:

      >>> pprint.pprint(dict(reader.get_object(
      ...     {'dict_data': [(1, '1'), (2, '2'), (3, '3')]},
      ...     None)))
      {1: '1', 2: '2', 3: '3'}

    It gets even worse with tuple keyed mappings,
    tuple keys get converted to JSON as a list,
    which need to be handled then by get_object

      >>> inp = {'dict_data': [([1, 'key-one', None], 'value-one'),
      ...                      ([2, 'key-two', True], 'value-two'),
      ...                      ('key-three', 'value-three')]}
      >>> pprint.pprint(dict(reader.get_object(inp, None)))
      {'key-three': 'value-three',
       (1, 'key-one', None): 'value-one',
       (2, 'key-two', True): 'value-two'}

    """

def doctest_ObjectReader_get_object_constant():
    """ObjectReader: get_object(): constant

      >>> reader = serialize.ObjectReader(dm)
      >>> reader.get_object(
      ...     {'_py_constant': 'pjpersist.tests.test_serialize.Constant'},
      ...     None)
      <pjpersist.tests.test_serialize.Constant object at ...>
      >>> reader.get_object(
      ...     {'_py_constant': 'pjpersist.interfaces.IObjectWriter'}, None)
      <InterfaceClass pjpersist.interfaces.IObjectWriter>
    """

def doctest_ObjectReader_get_ghost():
    """ObjectReader: get_ghost()

      >>> writer = serialize.ObjectWriter(dm)
      >>> top = Top()
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')

    The ghost object is a shell without any loaded object state:

      >>> reader = serialize.ObjectReader(dm)
      >>> gobj = reader.get_ghost(top._p_oid)
      >>> gobj._p_jar.__class__
      <class 'pjpersist.datamanager.PJDataManager'>
      >>> gobj._p_state
      0

    The second time we look up the object, it comes from cache:

      >>> gobj = reader.get_ghost(top._p_oid)
      >>> gobj._p_state
      0
    """

def doctest_ObjectReader_set_ghost_state():
    r"""ObjectReader: set_ghost_state()

      >>> writer = serialize.ObjectWriter(dm)
      >>> top = Top()
      >>> top.name = 'top'
      >>> writer.store(top)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test')

    The ghost object is a shell without any loaded object state:

      >>> reader = serialize.ObjectReader(dm)
      >>> gobj = reader.get_ghost(top._p_oid)
      >>> gobj._p_jar.__class__
      <class 'pjpersist.datamanager.PJDataManager'>
      >>> gobj._p_state
      0

    Now load the state:

      >>> reader.set_ghost_state(gobj)
      >>> gobj.name
      'top'

    """


def doctest_deserialize_persistent_references():
    """Deserialization o persistent references.

    The purpose of this test is to demonstrate the proper deserialization of
    persistent object references.

    Let's create a simple object hierarchy:

      >>> top = Top()
      >>> top.name = 'top'
      >>> top.foo = Foo()
      >>> top.foo.name = 'foo'

      >>> dm.root['top'] = top
      >>> commit()

    Let's check that the objects were properly serialized.

      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'foo': {'_py_type': 'DBREF',
                          'database': 'pjpersist_test',
                          'id': '0001020304050607080a0b0c0',
                          'table': 'Foo'},
                 'name': 'top'},
        'id': '0001020304050607080a0b0c0'}]

      >>> dumpTable('Foo')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Foo',
                 'name': 'foo'},
        'id': '0001020304050607080a0b0c0'}]

    Now we access the objects objects again to see whether they got properly
    deserialized.

      >>> top2 = dm.root['top']
      >>> id(top2) == id(top)
      False
      >>> top2.name
      'top'

      >>> id(top2.foo) == id(top.foo)
      False
      >>> top2.foo.__class__
      <class 'pjpersist.tests.test_serialize.Foo'>
      >>> top2.foo.name
      'foo'
    """


def doctest_deserialize_persistent_foreign_references():
    """
    Make sure we can reference objects from other databases.

    For this, we have to provide IPJDataManagerProvider

    First, store some object in one database
      >>> writer_other = serialize.ObjectWriter(dm_other)
      >>> top_other = Top()
      >>> top_other.name = 'top_other'
      >>> top_other.state = {'complex_data': 'value'}
      >>> writer_other.store(top_other)
      DBRef('Top', '0001020304050607080a0b0c', 'pjpersist_test_other')

    Store other object in datbase and refrence first one
      >>> writer_other = serialize.ObjectWriter(dm)
      >>> top = Top()
      >>> top.name = 'main'
      >>> top.other = top_other
      >>> dm.root['top'] = top
      >>> commit()

      >>> dumpTable('Top')
      [{'data': {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
                 'name': 'main',
                 'other': {'_py_type': 'DBREF',
                            'database': 'pjpersist_test_other',
                            'id': '0001020304050607080a0b0c0',
                            'table': 'Top'}},
        'id': '0001020304050607080a0b0c0'}]

      >>> top = dm.root['top']
      >>> print(top.name)
      main
      >>> print(top.other.name)
      top_other
      >>> top.other.state
      {'complex_data': 'value'}
    """


def doctest_PersistentDict_equality():
    """Test basic functions if PersistentDicts

      >>> import datetime
      >>> obj1 = serialize.PersistentDict({'key':'value'})
      >>> obj2 = serialize.PersistentDict({'key':'value'})
      >>> obj3 = serialize.PersistentDict({'key':None})
      >>> obj4 = serialize.PersistentDict({'key':datetime.datetime.now()})

      >>> obj1 == obj1 and obj2 == obj2 and obj3 == obj3 and obj4 == obj4
      True

      >>> obj1 == obj2
      True

      >>> obj1 == obj3
      False

      >>> obj1 == obj4
      False

      >>> obj3 == obj4
      False
    """


def doctest_table_decorator():
    """Test serialize.table

    This is our test class

      >>> @serialize.table('foobar_table')
      ... class Foo(object):
      ...     pass

    Check that TABLE_ATTR_NAME gets set

      >>> getattr(Foo, interfaces.TABLE_ATTR_NAME)
      'foobar_table'

    Check that TABLE_KLASS_MAP gets updated

      >>> len(serialize.TABLE_KLASS_MAP)
      1
      >>> list(serialize.TABLE_KLASS_MAP['foobar_table'])
      [<class '__main__.Foo'>]

    Add a few more classes

      >>> @serialize.table('barbar_table')
      ... class Bar(object):
      ...     pass

    Another typical case, base and subclass stored in the same table

      >>> @serialize.table('foobar_table')
      ... class FooFoo(Foo):
      ...     pass

    Dump TABLE_KLASS_MAP

      >>> pprint.pprint(
      ...     [(k, sorted(v, key=lambda cls:cls.__name__))
      ...      for k, v in sorted(serialize.TABLE_KLASS_MAP.items())])
      [('barbar_table', [<class '__main__.Bar'>]),
       ('foobar_table', [<class '__main__.Foo'>, <class '__main__.FooFoo'>])]

    Edge case, using the decorator on a non class fails:

      >>> serialize.table('foobar_table')(object())
      Traceback (most recent call last):
      ...
      TypeError: ("Can't declare _p_pj_table", <object object at ...>)

    """


class CustomDict(collections.UserDict, object):
    def __getstate__(self):
        return self.data

    def __setstate__(self, state):
        self.data = state


def doctest_non_persistent_object_empty_custom():
    """ObjectReader: get_non_persistent_object(): _py_persistent_type edge case

    When the state was empty __setstate__ did not get called.
    The result is such custom objects as `CustomDict` got broken.
    The `data` attribute was never set, thus `CustomDictInstance.items()`
    or any method call failed with
    `AttributeError: 'CustomDict' object has no attribute 'data'`

      >>> top = Top()
      >>> top2 = Top()

    Set an 'empty' dict

      >>> top2.sub = CustomDict()

      >>> list(top2.sub.items())
      []

      >>> top2._p_pj_sub_object = True
      >>> writer = serialize.ObjectWriter(dm)
      >>> state = writer.get_state(top2, top)

    Look at the state of `sub` here, it has just the class path, no data
    Because some optimizations remove data.
    
      >>> pprint.pprint(state)
      {'_py_persistent_type': 'pjpersist.tests.test_serialize.Top',
       'sub': {'_py_type': 'pjpersist.tests.test_serialize.CustomDict'}}

      >>> reader = serialize.ObjectReader(dm)
      >>> newobj = reader.get_non_persistent_object(state, None)
      >>> newobj.__class__
      <class 'pjpersist.tests.test_serialize.Top'>

    Make sure that the `CustomDict` object still works despite being 'empty'.
    IOW `__setstate__` was called

      >>> list(newobj.sub.items())
      []

    """


def test_suite():
    suite = doctest.DocTestSuite(
        setUp=testing.setUp, tearDown=testing.tearDown,
        checker=testing.checker,
        optionflags=testing.OPTIONFLAGS)
    suite.layer = testing.db_layer
    return suite




##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# Copyright (c) 2014 Shoobx, Inc.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import json
import pprint

from pjpersist import testing, mquery, sqlbuilder as sb

dataset = [
    {'foo': 'bar'},
    {'nr': 42},
    {'some': {'numbers': [42, 69, 105]}, "nr": None},
    {'more': {'strings': ['a', 'f', 'x']}, "nr": [1, None]},
    {'plan': 'getdown', 'day': 'Friday', 'drink': 'whiskey', 'nr': 42},
]

def jformat(obj):
    """Helper for inserting JSON"""
    return "('%s')" % json.dumps(obj).replace("'", "''")


def setUp(test):
    testing.setUp(test)

    conn = test.globs['conn']
    with conn.cursor() as cur:
        cur.execute("CREATE TABLE mq (id SERIAL PRIMARY KEY, data JSONB)")
        cur.execute("INSERT INTO mq (data) VALUES " +
                    ", ".join(jformat(datum) for datum in dataset))
    conn.commit()


def select(conn, query, print_sql=False):
    try:
        with conn.cursor() as cur:
            converter = mquery.Converter("mq", "data")
            sql = sb.sqlrepr(
                sb.Select(sb.Field("mq", "data"), where=converter.convert(query)),
                'postgres'
            )
            if print_sql:
                print('SQL> ', sql)
            cur.execute(sql)
            for e in cur.fetchall():
                pprint.pprint(e[0])
    finally:
        conn.rollback()

def doctest_operators():
    """Test simple selectors and comparisons:


    We can do simple matches of strings:

       >>> select(conn, {'foo': 'bar'})
       {'foo': 'bar'}

    And numbers:

       >>> select(conn, {'nr': 42})
       {'nr': 42}
       {'day': 'Friday', 'drink': 'whiskey', 'nr': 42, 'plan': 'getdown'}

    We can query for an element in the list:

       >>> select(conn, {'some.numbers': 69})
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

    Comparison operators:

       >>> select(conn, {'nr': {'$gt': 40}})
       {'nr': 42}
       {'more': {'strings': ['a', 'f', 'x']}, 'nr': [1, None]}
       {'day': 'Friday', 'drink': 'whiskey', 'nr': 42, 'plan': 'getdown'}

    List searches:

       >>> select(conn, {'nr': {'$in': [40, 41, 42]}})
       {'nr': 42}
       {'day': 'Friday', 'drink': 'whiskey', 'nr': 42, 'plan': 'getdown'}

       >>> select(conn, {'foo': {'$in': ['foo', 'bar', 'baz']}})
       {'foo': 'bar'}

       >>> select(conn, {'foo': {'$nin': ['foo', 'baz']}})
       {'foo': 'bar'}

       >>> select(conn, {'more.strings': {'$in': ['a', 'g']}})
       {'more': {'strings': ['a', 'f', 'x']}, 'nr': [1, None]}

       >>> select(conn, {'more.strings': {'$in': ['h', 'g']}})

       >>> select(conn, {'some.numbers': {'$in': [42, 200]}})
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

       >>> select(conn, {'some.numbers': {'$in': [80, 200]}})

    Edge cases with empty lists:

       >>> select(conn, {'nr': {'$in': []}}, True)
       SQL>  SELECT mq.data FROM mq WHERE 'f'

       >>> select(conn, {'nr': {'$nin': []}}, True)
       SQL>  SELECT mq.data FROM mq WHERE 't'
       {'foo': 'bar'}
       {'nr': 42}
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}
       {'more': {'strings': ['a', 'f', 'x']}, 'nr': [1, None]}
       {'day': 'Friday', 'drink': 'whiskey', 'nr': 42, 'plan': 'getdown'}


    Existence of keys:

       >>> select(conn, {'some.numbers': {'$exists': True}})
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

       >>> select(conn, {'nr': {'$exists': False}})
       {'foo': 'bar'}

    Negation of comparisons:

       >>> select(conn, {'nr': {'$not': {'$gt': 40}}})
       {'foo': 'bar'}
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

       >>> select(conn, {'nr': {'$not': {'$gt': 42}}})
       {'foo': 'bar'}
       {'nr': 42}
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}
       {'day': 'Friday', 'drink': 'whiskey', 'nr': 42, 'plan': 'getdown'}

    List sizes:

       >>> select(conn, {'some.numbers': {'$size': 3}})
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

    $all:

       >>> select(conn, {'some.numbers': {'$all': [69, 42]}})
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

       >>> select(conn, {'some.numbers': {'$all': [69, 43]}})

    $elemMatch matches when one of the array elements matches all
    conditions:

       >>> select(conn, {'some.numbers': {'$elemMatch':
       ...                   [{'$gt': 68}, {'$lt': 70}]}})
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}

       >>> select(conn, {'some.numbers': {'$elemMatch':
       ...                   [{'$gt': 60}, {'$lt': 65}]}})

    _id is specialcased:

       >>> select(conn, {'_id': 1}, True)
       SQL>  SELECT mq.data FROM mq WHERE ((mq.id) = (1))
       {'foo': 'bar'}

       >>> select(conn, {'_id': {'$lt': 2}}, True)
       SQL>  SELECT mq.data FROM mq WHERE ((mq.id) < (2))
       {'foo': 'bar'}

    Comparing to null in Mongo matches if the value is indeed null, or
    if the value does not exist, or ir the value is a list containing null!

       >>> select(conn, {'nr': None}, True)
       SQL>  SELECT mq.data FROM mq
               WHERE ((((mq.data) -> ('nr')) = ('null'))
                   OR ((('[null]'::jsonb) <@ ((mq.data) -> ('nr')))
                   OR (((mq.data) -> ('nr')) IS NULL)))
       {'foo': 'bar'}
       {'nr': None, 'some': {'numbers': [42, 69, 105]}}
       {'more': {'strings': ['a', 'f', 'x']}, 'nr': [1, None]}

    """


def test_suite():
    suite =  doctest.DocTestSuite(
        setUp=setUp, tearDown=testing.tearDown,
        checker=testing.checker,
        optionflags=testing.OPTIONFLAGS)
    suite.layer = testing.db_layer
    return suite



##############################################################################
#
# Copyright (c) 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PJ Persistence Performance Test"""
import logging
import optparse
import os
import persistent
import psycopg2
import random
import sys
import tempfile
import time
import transaction
import cProfile

from pjpersist import datamanager
from pjpersist import testing
from pjpersist import serialize
from pjpersist.zope import container

from pjpersist.tests import random_data

import zope.container
import zope.container.btree
import ZODB
import ZODB.FileStorage

PJLOGGER = logging.getLogger('pjpersist.table')


MULTIPLE_CLASSES = True
PROFILE = False
PROFILE_OUTPUT = '/tmp/cprofile'
LOG_SQL = False
BIGDATA = random_data.BIGDICT
#BIGDATA = random_data.HUGEDICT
CLEAR_CACHE = False


class People(container.PJContainer):
    _p_pj_table = 'people'
    _pj_table = 'person'
    _pj_mapping_key = 'name'

@serialize.table('address')
class Address(persistent.Persistent):
    #_p_pj_table = 'address'

    def __init__(self, city):
        self.city = city

@serialize.table('person')
class Person(persistent.Persistent, container.PJContained):
    #_p_pj_table = 'person'

    name = None
    age = None
    address = None
    data = None

    def __init__(self, name, age):
        self.name = name
        self.age = age
        self.address = Address('Boston %i' %age)
        # let's have some data for JSON
        if BIGDATA is not None:
            self.data = BIGDATA

    def __repr__(self):
        return '<%s %s @ %s [%s]>' %(
            self.__class__.__name__, self.name, self.age, self.__name__)

@serialize.table('person')
class Person2(Person):
    pass


class PerformanceBase(object):
    personKlass = None
    person2Klass = None
    profile_output = None

    def printResult(self, text, t1, t2, count=None):
        dur = t2-t1
        text += ':'
        ops = ''
        if count:
            ops = "%d ops/second" % (count / dur)

        print('%-25s %.4f secs %s' % (text, dur, ops))

        PJLOGGER.debug('=========== done: %s', text)

    def insertPeople(self, options):
        pass

    def getPeople(self, options):
        pass

    def slow_read(self, people, peopleCnt):
        # Profile slow read
        transaction.begin()
        t1 = time.time()
        if PROFILE:
            cProfile.runctx(
                '[people[name].name for name in people]', globals(), locals(),
                filename=self.profile_output+'_slow_read')
        else:
            [people[name].name for name in people]
        t2 = time.time()
        transaction.commit()
        self.printResult('Slow Read', t1, t2, peopleCnt)

    def read_list(self, people, peopleCnt):
        # Profile fast read (values)
        transaction.begin()
        t1 = time.time()
        if PROFILE:
            cProfile.runctx(
                '[p for p in list(people)]', globals(), locals(),
                filename=self.profile_output+'_read_list')
        else:
            [p for p in list(people)]
        t2 = time.time()
        transaction.commit()
        self.printResult('Read (list(keys()))', t1, t2, peopleCnt)

    def read_list_values(self, people, peopleCnt):
        # Profile fast read (values)
        transaction.begin()
        t1 = time.time()
        if PROFILE:
            cProfile.runctx(
                '[p for p in list(people.values())]', globals(), locals(),
                filename=self.profile_output+'_read_list_values')
        else:
            [p for p in list(people.values())]
        t2 = time.time()
        transaction.commit()
        self.printResult('Read (list(values()))', t1, t2, peopleCnt)

    def fast_read_values(self, people, peopleCnt):
        # Profile fast read (values)
        transaction.begin()
        t1 = time.time()
        if PROFILE:
            cProfile.runctx(
                '[person.name for person in people.values()]', globals(), locals(),
                filename=self.profile_output+'_fast_read_values')
        else:
            [person.name for person in people.values()]
        t2 = time.time()
        transaction.commit()
        self.printResult('Fast Read (values)', t1, t2, peopleCnt)

    def fast_read(self, people, peopleCnt):
        # Profile fast read
        transaction.begin()
        t1 = time.time()
        if PROFILE:
            cProfile.runctx(
                '[person.name for person in people.find()]', globals(), locals(),
                filename=self.profile_output+'_fast_read')
        else:
            [person.name for person in people.find()]
        t2 = time.time()
        transaction.commit()
        self.printResult('Fast Read (find)', t1, t2, peopleCnt)

    def object_caching(self, people, peopleCnt):
        # Profile object caching
        transaction.begin()
        # cache warmup
        [person.name for person in people.values()]
        t1 = time.time()
        [person.name for person in people.values()]
        #cProfile.runctx(
        #    '[person.name for person in people.values()]', globals(), locals())
        t2 = time.time()
        transaction.commit()
        self.printResult('Fast Read (caching x2)', t1, t2, peopleCnt*2)

        transaction.begin()
        # cache warmup
        [person.name for person in people.values()]
        t1 = time.time()
        [person.name for person in people.values()]
        [person.name for person in people.values()]
        #cProfile.runctx(
        #    '[person.name for person in people.values()]', globals(), locals())
        t2 = time.time()
        transaction.commit()
        self.printResult('Fast Read (caching x3)', t1, t2, peopleCnt*3)

    def modify(self, people, peopleCnt):
        # Profile modification
        t1 = time.time()
        def modify():
            for person in list(people.values()):
                person.name += 'X'
                person.age += 1
            transaction.commit()
        if PROFILE:
            cProfile.runctx(
                'modify()', globals(), locals(),
                filename=self.profile_output+'_modify')
        else:
            modify()
        t2 = time.time()
        self.printResult('Modification', t1, t2, peopleCnt)

    def delete(self, people, peopleCnt):
        # Profile deletion
        t1 = time.time()
        for name in people.keys():
            if PROFILE:
                cProfile.runctx(
                    'del people[name]', globals(), locals(),
                    filename=self.profile_output+'_delete')
            else:
                del people[name]
        transaction.commit()
        t2 = time.time()
        self.printResult('Deletion', t1, t2, peopleCnt)

    def run_basic_crud(self, options):
        people = self.insertPeople(options)

        peopleCnt = len(people)

        if CLEAR_CACHE:
            people = self.getPeople(options)
        self.slow_read(people, peopleCnt)
        if LOG_SQL:
            people._p_jar._report_stats()

        if CLEAR_CACHE:
            people = self.getPeople(options)
        self.read_list(people, peopleCnt)
        if LOG_SQL:
            people._p_jar._report_stats()

        if CLEAR_CACHE:
            people = self.getPeople(options)
        self.read_list_values(people, peopleCnt)
        if LOG_SQL:
            people._p_jar._report_stats()

        if CLEAR_CACHE:
            people = self.getPeople(options)
        self.fast_read_values(people, peopleCnt)
        if LOG_SQL:
            people._p_jar._report_stats()

        if CLEAR_CACHE:
            people = self.getPeople(options)
        self.fast_read(people, peopleCnt)
        if LOG_SQL:
            people._p_jar._report_stats()

        if CLEAR_CACHE:
            people = self.getPeople(options)
        self.object_caching(people, peopleCnt)
        if LOG_SQL:
            people._p_jar._report_stats()

        if options.modify:
            if CLEAR_CACHE:
                people = self.getPeople(options)
            self.modify(people, peopleCnt)
            if LOG_SQL:
                people._p_jar._report_stats()

        if options.delete:
            if CLEAR_CACHE:
                people = self.getPeople(options)
            self.delete(people, peopleCnt)
            if LOG_SQL:
                people._p_jar._report_stats()


def getConnection(database=None):
    return psycopg2.connect(
        database=database or 'template1',
        host='localhost', port=5432,
        user='pjpersist', password='pjpersist')


class PerformancePJ(PerformanceBase):
    personKlass = Person
    person2Klass = Person2

    profile_output = PROFILE_OUTPUT + '_pj_'

    def insertPeople(self, options):
        if options.reload:
            connroot = getConnection()
            with connroot.cursor() as cur:
                cur.execute('END;')
                cur.execute('DROP DATABASE IF EXISTS performance;')
                cur.execute('CREATE DATABASE performance;')
            connroot.commit()

            datamanager.PJ_AUTO_CREATE_TABLES = False
            if LOG_SQL:
                testing.log_sql_to_file('/tmp/pjpersist.table.log')

            conn = getConnection('performance')

            dm = datamanager.PJDataManager(conn)

            dm.create_tables(('people', 'person', 'address'))
            dm.root._init_table()
            #dm._new_obj_cache._ensure_db_objects()

            # this speeds up slow_read around TWICE
            #with conn.cursor() as cur:
            #    cur.execute('END;')
            #    cur.execute(
            #        "CREATE INDEX data_name ON person ((data->>('name')));")

            dm.root['people'] = people = People()
            transaction.commit()

            def insert():
                for idx in xrange(options.size):
                    klass = (self.personKlass if (MULTIPLE_CLASSES and idx % 2)
                             else self.person2Klass)
                    people[None] = klass('Mr Number %.5i' % idx,
                                         random.randint(0, 100))

            # Profile inserts
            t1 = time.time()
            if PROFILE:
                cProfile.runctx(
                    'insert()', globals(), locals(),
                    filename=self.profile_output+'_insert')
            else:
                insert()
            transaction.commit()
            t2 = time.time()
            self.printResult('Insert', t1, t2, options.size)
        else:
            people = dm.root['people']

        return people

    def getPeople(self, options):
        conn = getConnection('performance')

        dm = datamanager.PJDataManager(conn)
        people = dm.root['people']
        return people


class PeopleZ(zope.container.btree.BTreeContainer):
    pass

class AddressZ(persistent.Persistent):

    def __init__(self, city):
        self.city = city

class PersonZ(persistent.Persistent, zope.container.contained.Contained):

    def __init__(self, name, age):
        self.name = name
        self.age = age
        self.address = AddressZ('Boston %i' %age)

    def __repr__(self):
        return '<%s %s @ %i [%s]>' %(
            self.__class__.__name__, self.name, self.age, self.__name__)

class Person2Z(Person):
    pass


class PerformanceZODB(PerformanceBase):
    personKlass = PersonZ
    person2Klass = Person2Z

    profile_output = PROFILE_OUTPUT + '_zodb_'

    def insertPeople(self, options):
        folder = tempfile.gettempdir()
        #folder = './'  # my /tmp is a tmpfs
        fname = os.path.join(folder, 'performance_data.fs')
        if options.reload:
            try:
                os.remove(fname)
            except:
                pass
        fs = ZODB.FileStorage.FileStorage(fname)
        self.db = ZODB.DB(fs)
        conn = self.db.open()

        root = conn.root()

        if options.reload:
            root['people'] = people = PeopleZ()
            transaction.commit()

            # Profile inserts
            transaction.begin()
            t1 = time.time()
            for idx in xrange(options.size):
                klass = (self.personKlass if (MULTIPLE_CLASSES and idx % 2)
                         else self.person2Klass)
                name = 'Mr Number %.5i' % idx
                people[name] = klass(name, random.randint(0, 100))
            transaction.commit()
            t2 = time.time()
            self.printResult('Insert', t1, t2, options.size)
        else:
            people = root['people']

        return people

    def getPeople(self, options):
        conn = self.db.open()

        root = conn.root()
        people = root['people']

        return people

    def fast_read(self, people, peopleCnt):
        pass


parser = optparse.OptionParser()
parser.usage = '%prog [options]'

parser.add_option(
    '-s', '--size', action='store', type='int',
    dest='size', default=1000,
    help='The amount of objects to use.')

parser.add_option(
    '--no-reload', action='store_false',
    dest='reload', default=True,
    help='A flag, when set, causes the DB not to be reloaded.')

parser.add_option(
    '--no-modify', action='store_false',
    dest='modify', default=True,
    help='A flag, when set, causes the data not to be modified.')

parser.add_option(
    '--no-delete', action='store_false',
    dest='delete', default=True,
    help='A flag, when set, causes the data not to be deleted at the end.')


def main(args=None):
    # Parse command line options.
    if args is None:
        args = sys.argv[1:]
    options, args = parser.parse_args(args)

    testing.setUpSerializers(None)

    print('PJ ---------------')
    PerformancePJ().run_basic_crud(options)
    print('ZODB  ---------------')
    PerformanceZODB().run_basic_crud(options)
